"""Display a PIV field (:mod:`fluidimage.data_objects.display_piv`)
===================================================================

.. autoclass:: DisplayPIV
   :members:
   :private-members:

"""

import matplotlib.pyplot as plt
import numpy as np

from ..calcul.correl import compute_indices_from_displacement


class DisplayPIV:
    """Display a piv result object."""

    def __init__(
        self,
        im0,
        im1,
        piv_results=None,
        show_interp=False,
        scale=0.2,
        show_error=True,
        pourcent_histo=99,
        hist=False,
        show_correl=True,
    ):

        self.piv_results = piv_results

        if (
            show_correl
            and hasattr(piv_results, "correls")
            and piv_results.correls
        ):
            self.show_correl = True
        else:
            self.show_correl = False

        fig = plt.figure()
        fig.event_handler = self

        if self.show_correl:
            ax1 = plt.subplot(121)
            ax2 = plt.subplot(122)
            self.ax2 = ax2
        else:
            ax1 = plt.gca()

        self.fig = fig
        self.ax1 = ax1

        if im0 is not None:
            p0 = np.percentile(
                np.reshape(im0, (1, np.product(im0.shape))).transpose(),
                pourcent_histo,
            )
            p1 = np.percentile(
                np.reshape(im1, (1, np.product(im1.shape))).transpose(),
                pourcent_histo,
            )

            im0 = im0.copy()
            im1 = im1.copy()

            im0[im0 > p0] = p0
            im1[im1 > p1] = p1

            self.image0 = ax1.imshow(
                im0,
                interpolation="nearest",
                cmap=plt.cm.gray,
                origin="upper",
                extent=[0, im0.shape[1], im0.shape[0], 0],
                vmin=0,
                vmax=0.99 * im0.max(),
            )

            self.image1 = ax1.imshow(
                im1,
                interpolation="nearest",
                cmap=plt.cm.gray,
                origin="upper",
                extent=[0, im0.shape[1], im0.shape[0], 0],
                vmin=0,
                vmax=0.99 * im1.max(),
            )
            self.image1.set_visible(False)
        else:
            self.image0 = None

        (l,) = ax1.plot(0, 0, "oy")
        l.set_visible(False)

        ax1.set_title("im 0 (alt+s to switch)")

        t = fig.text(0.1, 0.05, "")

        self.t = t
        self.l = l

        if im0 is not None:
            ax1.set_xlim(0, im0.shape[1])
            ax1.set_ylim(im0.shape[0], 0)

        ax1.set_xlabel("pixels")
        ax1.set_ylabel("pixels")

        if piv_results is not None:
            if show_interp:
                if hasattr(piv_results, "deltaxs_approx"):
                    deltaxs = piv_results.deltaxs_approx
                    deltays = piv_results.deltays_approx
                    xs = piv_results.ixvecs_approx
                    ys = piv_results.iyvecs_approx
                else:
                    deltaxs = piv_results.deltaxs_final
                    deltays = piv_results.deltays_final
                    xs = piv_results.ixvecs_final
                    ys = piv_results.iyvecs_final
            else:
                deltaxs = piv_results.deltaxs
                deltays = piv_results.deltays
                xs = piv_results.xs
                ys = piv_results.ys

            if im0 is None:
                deltays *= -1

            self.q = ax1.quiver(
                xs,
                ys,
                deltaxs,
                -deltays,
                width=0.004,
                picker=10,
                color="c",
                scale_units="xy",
                scale=scale,
            )

            if show_error and not hasattr(piv_results, "deltays_wrong"):
                show_error = False

            if show_error:
                self.inds_error = inds_error = np.array(
                    list(piv_results.deltays_wrong.keys()), dtype=int
                )
                xs_wrong = xs[inds_error]
                ys_wrong = ys[inds_error]
                dxs_wrong = np.array(
                    [piv_results.deltaxs_wrong[i] for i in inds_error]
                )
                dys_wrong = np.array(
                    [piv_results.deltays_wrong[i] for i in inds_error]
                )

                if im0 is None:
                    dys_wrong *= -1

                self.q_wrong = ax1.quiver(
                    xs_wrong,
                    ys_wrong,
                    dxs_wrong,
                    -dys_wrong,
                    picker=10,
                    color="r",
                    scale_units="xy",
                    scale=scale,
                )

                inds_isnan = inds_error[np.isnan(dxs_wrong)]
                self.inds_isnan = inds_isnan
                xs_isnan = xs[inds_isnan]
                ys_isnan = ys[inds_isnan]

                zeros = np.zeros_like(xs_isnan)
                self.q_isnan = ax1.quiver(
                    xs_isnan,
                    ys_isnan,
                    zeros,
                    zeros,
                    minshaft=4,
                    picker=10,
                    color="r",
                    scale_units="xy",
                    scale=scale,
                )

        if hist:
            fig2, axes = plt.subplots(ncols=2)
            ax3, ax4 = axes.ravel()
            ind = (
                np.isnan(deltaxs)
                + np.isnan(deltays)
                + np.isinf(deltaxs)
                + np.isinf(deltays)
            )
            deltaxs2 = deltaxs[~ind]
            deltays2 = deltays[~ind]
            ax3.hist(deltaxs2, "fd", color="b", label=r"$\Delta x_s$")
            ax3.hist(deltays2, "fd", color="r", label=r"$\Delta y_s$")
            ax3.set_xlabel("displacement x (blue) and y (red) (pixels)")
            ax3.set_ylabel("histogram")
            ax3.legend()

            ax4.hist(piv_results.correls_max, "fd", color="g")
            ax4.set_xlabel("Maximum pixel correlation")
            ax4.set_ylabel("histogram")
            fig2.show()

        self.ind = 0
        fig.canvas.mpl_connect("pick_event", self.onpick)
        fig.canvas.mpl_connect("key_press_event", self.onclick)

        print("press alt+h for help")

        plt.show()

    def onclick(self, event):
        if event.key == "alt+h":
            print(
                "\nclick on a vector to show information\n"
                "alt+s\t\t\t switch between images\n"
                "alt+left or alt+right\t change vector."
            )

        if event.inaxes != self.ax1:
            return

        if event.key == "alt+s":
            self.switch()

        if event.key == "alt+left":
            self.select_arrow(self.ind - 1)

        if event.key == "alt+right":
            self.select_arrow(self.ind + 1)

    def onpick(self, event):
        if not (
            event.artist == self.q
            or event.artist == self.q_wrong
            or event.artist == self.q_isnan
        ):
            return True

        # the click locations
        # x = event.mouseevent.xdata
        # y = event.mouseevent.ydata
        ind = event.ind
        self.select_arrow(ind, event.artist)

    def select_arrow(self, ind, artist=None):
        try:
            ind = ind[0]
        except (TypeError, IndexError):
            return

        if artist is None:
            print("artist is None")
            return

        # if ind in self.piv_results.errors.keys():
        #     artist = self.q_wrong
        #     ind = self.inds_error.index(ind)
        # else:
        #     artist = self.q

        if artist == self.q:
            ind_all = ind
            q = self.q
        elif artist == self.q_wrong:
            ind_all = self.inds_error[ind]
            q = self.q_wrong
        elif artist == self.q_isnan:
            ind_all = self.inds_isnan[ind]
            q = self.q_isnan
        else:
            raise NotImplementedError("other artist" + str(artist))

        if ind >= len(q.X) or ind < 0:
            return

        self.ind = ind_all

        result = self.piv_results

        ix = q.X[ind]
        iy = q.Y[ind]
        deltax = result.deltaxs[ind_all]
        deltay = result.deltays[ind_all]

        if np.isnan(deltax):
            deltax = result.deltaxs_wrong[ind_all]
            deltay = result.deltays_wrong[ind_all]

        self.l.set_visible(True)
        self.l.set_data(ix, iy)

        text = (
            f"vector at ix = {ix} : iy = {iy}"
            f" ; U = {deltax:.3f} ; V = {deltay:.3f}"
        )

        if result.correls_max is not None:
            correl_max = result.correls_max[ind_all]
            text += f", C = {correl_max:.3f}"

        if (
            self.piv_results.errors is not None
            and ind_all in self.piv_results.errors
        ):
            text += ", error: " + self.piv_results.errors[ind_all]

        self.t.set_text(text)

        if self.show_correl:
            ax2 = self.ax2
            ax2.cla()
            alphac = result.correls[ind_all]
            alphac_max = alphac.max()
            correl = correl_max / alphac_max * alphac

            ax2.imshow(correl, origin="lower", interpolation="none", vmin=0)

            ax2.plot(
                result.indices_no_displacement[0],
                result.indices_no_displacement[1],
                "or",
            )

            try:
                deltax -= result.deltaxs_approx0[ind_all]
                deltay -= result.deltays_approx0[ind_all]
            except AttributeError:
                pass

            i1, i0 = compute_indices_from_displacement(
                deltax, deltay, result.indices_no_displacement
            )

            ax2.plot(i1, i0, "xr")

            if hasattr(result, "deltaxs_final"):
                deltax = result.deltaxs_final[ind_all]
                deltay = result.deltays_final[ind_all]

                try:
                    deltax -= result.deltaxs_approx0[ind_all]
                    deltay -= result.deltays_approx0[ind_all]
                except AttributeError:
                    pass

                i1, i0 = compute_indices_from_displacement(
                    deltax, deltay, result.indices_no_displacement
                )

                ax2.plot(i1, i0, "ow")

            params = self.piv_results.params

            if params.piv0.nb_peaks_to_search > 1:
                other_peaks = result.secondary_peaks[ind_all]
                if other_peaks is not None:
                    if len(other_peaks) == 0:
                        s = "no other peak"
                        ax2.set_title(s)
                        print(s)
                    else:
                        s = "{} other peaks".format(len(other_peaks))
                        ax2.set_title(s)
                        print(s)

                    for (dx, dy, cmax) in other_peaks:
                        i1, i0 = compute_indices_from_displacement(
                            dx, dy, result.indices_no_displacement
                        )
                        ax2.plot(i1, i0, "sr")
                        print(dx, dy, cmax)

            if params.piv0.displacement_max is not None:
                circle = plt.Circle(
                    result.indices_no_displacement,
                    result.displacement_max,
                    color="b",
                    fill=False,
                )
                ax2.add_artist(circle)

            ax2.axis("scaled")
            ax2.set_xlim(-0.5, correl.shape[1] - 0.5)
            ax2.set_ylim(-0.5, correl.shape[0] - 0.5)
        self.fig.canvas.draw()

    def switch(self):
        if self.image0 is not None:
            self.image0.set_visible(not self.image0.get_visible())
            self.image1.set_visible(not self.image1.get_visible())

            self.ax1.set_title(
                "im {} (alt+s to switch)".format(int(self.image1.get_visible()))
            )

            self.fig.canvas.draw()



import numpy as np

from .util import compute_1dspectrum, compute_2dspectrum, reshape_on_grid_final


def test_compute_1dspectrum():
    nx = 100
    lx = 10.0
    dx = lx / nx
    x = dx * np.arange(nx)
    signal = np.random.rand(nx)

    energy = np.mean(signal**2) / 2

    signal_fft, omega, psd = compute_1dspectrum(x, signal)

    energy_fft = 0.5 * np.sum(np.abs(signal_fft) ** 2)
    assert np.allclose(energy, energy_fft), (energy, energy_fft)
    domega = omega[1] - omega[0]
    assert np.allclose(domega, 2 * np.pi / lx)
    assert np.allclose(energy, np.sum(psd) * domega)


def _test_compute_2dspectrum(nt):

    nx = 10
    ny = 100
    lx = 1000
    ly = 10000
    dx = lx / nx
    dy = ly / ny
    x = dx * np.arange(nx)
    y = dy * np.arange(ny)

    X, Y = np.meshgrid(x, y)

    assert X.shape == (ny, nx)

    nt = None

    if nt:
        axes = (1, 2)
        size = X.size * nt
        shape = [nt, *X.shape]
    else:
        nt = 1
        axes = (0, 1)
        size = X.size
        shape = X.shape

    signal = np.random.rand(size).reshape(shape)

    energy = 0.5 * np.mean(signal**2)

    signal_fft, kx, ky, psd = compute_2dspectrum(X, Y, signal, axes=axes)
    energy_fft = 0.5 * np.sum(np.abs(signal_fft) ** 2) / nt
    assert np.allclose(energy, energy_fft), (energy, energy_fft)
    dkx = kx[1] - kx[0]
    dky = ky[1] - ky[0]

    energy_psd = dkx * dky * np.sum(psd) / nt
    assert np.allclose(energy, energy_psd)


def test_compute_2dspectrum_input2d():
    _test_compute_2dspectrum(nt=None)


def test_compute_2dspectrum_input3d():
    _test_compute_2dspectrum(nt=4)


def test_reshape_on_grid_final():

    x_increasing = np.linspace(-1, 1, 2)
    x_decreasing = np.linspace(2, 1, 2)
    y_increasing = np.linspace(0, 2, 10)
    y_decreasing = np.linspace(2, 0, 10)

    x_grid_ii, y_grid_ii = np.meshgrid(x_increasing, y_increasing)
    x_grid_id, y_grid_id = np.meshgrid(x_increasing, y_decreasing)
    x_grid_dd, y_grid_dd = np.meshgrid(x_decreasing, y_decreasing)

    x_flat_ii, y_flat_ii = x_grid_ii.flatten(), y_grid_ii.flatten()
    x_flat_id, y_flat_id = x_grid_id.flatten(), y_grid_id.flatten()
    x_flat_dd, y_flat_dd = x_grid_dd.flatten(), y_grid_dd.flatten()

    out_x_grid_ii, out_y_grid_ii, plus_ii, minus_ii = reshape_on_grid_final(
        x_flat_ii, y_flat_ii, x_flat_ii + y_flat_ii, x_flat_ii - y_flat_ii
    )

    out_x_grid_id, out_y_grid_id, plus_id, minus_id = reshape_on_grid_final(
        x_flat_id, y_flat_id, x_flat_id + y_flat_id, x_flat_id - y_flat_id
    )

    out_x_grid_dd, out_y_grid_dd, plus_dd, minus_dd = reshape_on_grid_final(
        x_flat_dd, y_flat_dd, x_flat_dd + y_flat_dd, x_flat_dd - y_flat_dd
    )

    assert np.allclose(out_x_grid_ii, x_grid_ii)
    assert np.allclose(out_y_grid_ii, y_grid_ii)
    assert np.allclose(out_x_grid_id, x_grid_id)
    assert np.allclose(out_y_grid_id, y_grid_id)
    assert np.allclose(out_x_grid_dd, x_grid_dd)
    assert np.allclose(out_y_grid_dd, y_grid_dd)

    for iy in range(y_increasing.size):
        for ix in range(x_increasing.size):
            assert (
                out_x_grid_ii[iy, ix] + out_y_grid_ii[iy, ix] == plus_ii[iy, ix]
            )
            assert (
                out_x_grid_ii[iy, ix] - out_y_grid_ii[iy, ix] == minus_ii[iy, ix]
            )
            assert (
                out_x_grid_id[iy, ix] + out_y_grid_id[iy, ix] == plus_id[iy, ix]
            )
            assert (
                out_x_grid_id[iy, ix] - out_y_grid_id[iy, ix] == minus_id[iy, ix]
            )
            assert (
                out_x_grid_dd[iy, ix] + out_y_grid_dd[iy, ix] == plus_dd[iy, ix]
            )
            assert (
                out_x_grid_dd[iy, ix] - out_y_grid_dd[iy, ix] == minus_dd[iy, ix]
            )



"""
Module where the version is written.

It is executed in setup.py and imported in fluidimage/__init__.py.

See:

http://en.wikipedia.org/wiki/Software_versioning
http://legacy.python.org/dev/peps/pep-0386/

'a' or 'alpha' means alpha version (internal testing),
'b' or 'beta' means beta version (external testing).
"""
__version__ = "0.1.4"

try:
    from fluidimage._hg_rev import hg_rev
except ImportError:
    hg_rev = "?"



"""Utilities for post-processing (:mod:`fluidimage.postproc.util`)
==================================================================

.. autofunction:: get_grid_from_ivecs_final

.. autofunction:: reshape_on_grid_final

.. autofunction:: compute_rot

.. autofunction:: compute_div

.. autofunction:: compute_1dspectrum

.. autofunction:: compute_2dspectrum
"""

import numpy as np


def get_grid_from_ivecs_final(x_flat, y_flat):
    """Get a 2d grid from flat arrays"""

    x = np.unique(x_flat)
    y = np.unique(y_flat)

    if x_flat[1] == x_flat[0]:
        assert y_flat[1] != y_flat[0]
        # second_index_corresponds_to_x = True
        dx = x_flat[len(y)] - x_flat[0]
        dy = y_flat[1] - y_flat[0]
    else:
        assert y_flat[1] == y_flat[0]
        # second_index_corresponds_to_x = False
        dx = x_flat[1] - x_flat[0]
        dy = y_flat[len(x)] - y_flat[0]

    if dx < 0:
        x = x[::-1]

    if dy < 0:
        y = y[::-1]

    X, Y = np.meshgrid(x, y)
    return X, Y


def reshape_on_grid_final(x_flat, y_flat, deltaxs, deltays):
    """Reshape flat arrays on a 2d grid"""

    X, Y = get_grid_from_ivecs_final(x_flat, y_flat)
    shape = X.shape
    if x_flat[1] == x_flat[0]:
        assert y_flat[1] != y_flat[0]
        second_index_corresponds_to_x = True
        shape = shape[::-1]
    else:
        assert y_flat[1] == y_flat[0]
        second_index_corresponds_to_x = False

    U = np.reshape(deltaxs, shape)
    V = np.reshape(deltays, shape)

    if second_index_corresponds_to_x:
        U = U.T
        V = V.T

    return X, Y, U, V


def compute_rot(dUdy, dVdx):
    """Compute the rotational"""
    return dVdx - dUdy


def compute_div(dUdx, dVdy):
    """Compute the divergence"""
    return dUdx + dVdy


def compute_1dspectrum(x, signal, axis=0):

    """
    Computes the 1D Fourier Transform

    Parameters
    ----------

    x: 1D np.ndarray

    signal: np.ndarray

    axis: int

      Direction of the Fourier transform

    Returns
    -------

    signal_fft: np.ndarray

      fourier transform

    omega: np.ndarray

      puslation (in rad/s if x in s)

    psd: np.ndarray

      power spectral density normalized such that
      np.sum(signal**2) * dx / Lx = np.sum(psd) * domega

    """
    n = x.size
    dx = x[1] - x[0]
    Lx = n * dx
    dk = 2 * np.pi / Lx

    signal_fft = (
        dx / Lx * np.fft.fftshift(np.fft.fft(signal, axis=axis), axes=axis)
    )
    omega = np.fft.fftshift(np.fft.fftfreq(n, dx)) * 2 * np.pi

    psd = 0.5 / dk * np.abs(signal_fft) ** 2

    return signal_fft, omega, psd


def compute_2dspectrum(X, Y, signal, axes=(1, 2)):
    """
    Computes the 2D Fourier Transform

    INPUT:
    X: 2D np.array
    Y: 2D np.array
    signal: np.array to Fourier transform
    axis: directions of the fourier transform

    OUTPUT:
    signal_fft = fourier transform
    kx: puslation (in rad/m if X in m)
    ky: pulsation (in rad/m if Y in m)
    psd: power spectral density normalized such that
    np.sum(signal**2) * dx / Lx = np.sum(psd) * domega
    """
    if X.ndim == 2:
        nx = X.shape[1]
        ny = X.shape[0]
        x = X[0, :2]
        y = Y[:2, 0]
    else:
        nx = X.size
        x = X
        ny = Y.size
        y = Y

    dx = x[1] - x[0]
    dy = y[1] - y[0]
    lx = nx * dx
    ly = ny * dy
    dkx = 2 * np.pi / lx
    dky = 2 * np.pi / ly

    # energy = 0.5 * np.mean(signal**2)

    signal_fft = (
        1 / (nx * ny) * np.fft.fftshift(np.fft.fft2(signal, axes=axes), axes=axes)
    )

    # if axes == (1, 2):
    #     nt = signal.shape[0]
    # else:
    #     nt = 1

    # energy_fft = 0.5 / nt * np.sum(np.abs(signal_fft)**2)
    # assert np.allclose(energy, energy_fft), (energy, energy_fft)

    # in rad/m
    kx = np.fft.fftshift(np.fft.fftfreq(nx, dx)) * (2 * np.pi)
    ky = np.fft.fftshift(np.fft.fftfreq(ny, dy)) * (2 * np.pi)

    psd = 0.5 / (dkx * dky) * np.abs(signal_fft) ** 2

    # energy_psd = dkx * dky * np.sum(psd) / nt
    # assert np.allclose(energy, energy_psd)

    return signal_fft, kx, ky, psd



import matplotlib.pyplot as plt
import numpy as np

from fluidimage.calcul.correl import (
    CorrelFFTNumpy,
    CorrelFFTW,
    CorrelScipyNdimage,
    CorrelScipySignal,
)
from fluidimage.synthetic import make_synthetic_images

plt.ion()


nx = 32
ny = 32
displacement_x = 2.0
displacement_y = 2.0

displacements = np.array([displacement_x, displacement_y])

nb_particles = (nx // 3) ** 2


print(f"nx: {nx} ; ny: {ny}")

im0, im1 = make_synthetic_images(
    displacements, nb_particles, shape_im0=(ny, nx), epsilon=0.0
)

# plt.figure()

# ax0 = plt.subplot(121)
# ax1 = plt.subplot(122)

# axi0 = ax0.imshow(im0, interpolation='nearest')
# axi1 = ax1.imshow(im1, interpolation='nearest')


classes = {
    "sig": CorrelScipySignal,
    "ndimage": CorrelScipyNdimage,
    "np.fft": CorrelFFTNumpy,
    "fftw": CorrelFFTW,
}


cs = {}
correls = {}
for k, cls in classes.items():
    print(k, cls)
    correl = cls(im0.shape, im1.shape)
    correls[k] = correl
    c, norm = correl(im0, im1)
    cs[k] = c

    inds_max = np.array(np.unravel_index(c.argmax(), c.shape))
    print(inds_max)

    displacement_computed = correl.compute_displacement_from_indices(inds_max)
    print(displacements.astype("int"), displacement_computed)

    if not np.allclose(displacements.astype("int"), displacement_computed):
        print(
            "do not understand " + k,
            displacements.astype("int"),
            displacement_computed,
        )


# plt.figure()

# i = 1
# for k, c in cs.items():
#     ax1 = plt.subplot(2, 2, i)
#     ax1.imshow(c, interpolation='nearest')
#     i += 1

# plt.show()



"""Surface tracking (:mod:`fluidimage.works.surface_tracking`)
==============================================================

.. autoclass:: WorkSurfaceTracking
   :members:
   :private-members:

"""
###############################################################################
# !/usr/bin/env python                                                        #
#  -*- coding: utf-8 -*-                                                      #
#                         (C) Cyrille Bonamy, Stefan Hoerner, 2017            #
#            LEGI Grenoble, University Otto-von-Guericke Magdeburg            #
###############################################################################
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                        #
# See the GNU General Public License for more details.                        #
# You should have received a copy of the GNU General Public License           #
# along with this program.                                                    #
# If not, see <http://www.gnu.org/licenses/>.                                 #
###############################################################################
#     This function provides surface tracking tools, it is part of the        #
#            oscillating profile experiment at LEGI 2017                      #
###############################################################################

import math
import sys
from pathlib import Path

import numpy as np
import scipy.interpolate
import scipy.io
from skimage.transform import resize

from fluiddyn.util.paramcontainer import ParamContainer
from fluidimage import SerieOfArraysFromFiles
from fluidimage.util import imread, logger

from . import BaseWork

# import matplotlib.pyplot as plt


class WorkSurfaceTracking(BaseWork):
    """Main work for surface tracking

    Parameters
    ----------

    params : :class:`fluiddyn.util.paramcontainer.ParamContainer`

      The default parameters are obtained from the class method
      :func:`WorkSurfaceTracking.create_default_params`.

    """

    @classmethod
    def create_default_params(cls):
        "Create an object containing the default parameters (class method)."
        params = ParamContainer(tag="params")
        cls._complete_params_with_default(params)
        return params

    @classmethod
    def _complete_params_with_default(cls, params):
        params._set_child(
            "surface_tracking",
            attribs={
                "xmin": 475,
                "xmax": 640,
                "ymin": 50,
                "ymax": 700,
                "distance_lens": 0.36,
                "distance_object": 1.293,
                "pix_size": 2.4 * 10**-4,
                "startref_frame": 0,
                "lastref_frame": 49,
                "sur": 16,
                "k_x": 40.625,
                "k_y": 0,
                "slicer": 4,
                "red_factor": 1,
                "n_frames_stock": 1,
                "crop_edge": False,
                "borders": 1,
                "correct_pos": False,
                "correct_height": False,
                "offset": 0.0,
            },
        )

        params.surface_tracking._set_doc(
            """Surface Tracking parameters:

xmin: int (default 475)
    xmin to crop the image im[xmin:xmax, ymin:ymax].

xmax: int (default 640)
    xmax to crop the image im[xmin:xmax, ymin:ymax]

ymin: int (default 50)
    ymin to crop the image im[xmin:xmax, ymin:ymax]

ymax: int (default 700)
    ymax to crop the image im[xmin:xmax, ymin:ymax]

distance_lens: float (default 0.36)
    distance in [m] lenses of camera/projetor

distance_object: float (default 1.07)
    distance in [m] camera/projector and surface

pix_size: float (default 2.4 * 10 ** -4)
    pixel size

startref_frame: int (default 0)
    indice of first reference image

lastref_frame: int (default 49)
    indice of last reference image

sur: int (default 16)

k_x: float (default 70.75)
    wave vector oj. grid (approx. value, will set accurate later)

k_y: float (default 0)
    wave vector of the grid y-axis

slicer: int (default 4)
    cut the borders

red_factor: int (default 1)
    reduction factor to for the pixels to take tp speed up

n_frames_stock: int (default 1)
    number of frames to stock in one file

crop_edge: boolean (default False)
    searches for the structure and crops the part of the frame outside of
    the structure

borders: int (default 7)
    pixel to set zero height additional to the borders if the structure was
    cropped to avoid jerks

correct_pos: boolean (default=False)
    correct position of the height (necessary for large heights)

correct_height: boolean (default=False)
    correct height by a reference (provided from path_ref)

offset: float (default 0.0)
    height of the reference surface in [m] to zero level
"""
        )

    def __init__(self, params):

        self.cpt = 0

        self.params = params

        self.works_surface_tracking = []
        self.nameFrame = None

        self.path = params.images.path
        self.path_ref = params.images.path_ref

        self.verify_process = False
        self.ref_film = None
        self.filmName = None
        self.save_png = True
        self.thres = 300
        self.crop_edge = self.params.surface_tracking.crop_edge
        self.borders = self.params.surface_tracking.borders
        self.correct_pos = self.params.surface_tracking.correct_pos
        self.correct_height = self.params.surface_tracking.correct_height
        self.offset = self.params.surface_tracking.offset
        self.xmin = self.params.surface_tracking.xmin
        self.xmax = self.params.surface_tracking.xmax
        self.ymin = self.params.surface_tracking.ymin
        self.ymax = self.params.surface_tracking.ymax

        self.distance_lens = self.params.surface_tracking.distance_lens
        self.distance_object = self.params.surface_tracking.distance_object
        self.pix_size = self.params.surface_tracking.pix_size

        self.startref_frame = self.params.surface_tracking.startref_frame
        self.lastref_frame = self.params.surface_tracking.lastref_frame
        self.sur = self.params.surface_tracking.sur

        self.k_y = self.params.surface_tracking.k_y
        self.slicer = self.params.surface_tracking.slicer

        self.red_factor = self.params.surface_tracking.red_factor
        self.n_frames_stock = self.params.surface_tracking.n_frames_stock

        self.plot_reduction_factor = 10
        self.l_x = self.xmax - self.xmin
        self.l_y = self.ymax - self.ymin

        # wave_proj_pix = self.wave_proj / self.pix_size

        self.kx = np.arange(-self.l_x / 2, self.l_x / 2) / self.l_x
        self.ky = np.arange(-self.l_y / 2, self.l_y / 2) / self.l_y

        self.refserie = SerieOfArraysFromFiles(
            params.images.path_ref, params.images.str_slice_ref
        )
        k_x = self.compute_kx(self.refserie)
        logger.info("Value of kx computed = " + str(k_x))
        self.kslicer = 2 * k_x
        self.wave_proj = 1 / (k_x / self.l_x / self.pix_size)
        self.kxx = self.kx / self.pix_size
        self.gain, self.filt = self.set_gain_filter(
            k_x, self.l_y, self.l_x, self.slicer
        )
        self.a1_tmp = None
        self.ref_height = self.process_ref()
        logger.info("reference computed")

    def compute_kx(self, serie):
        """calculates the average wave vector from a set of reference images

        Parameters
        ----------
        series: int
        set of reference frames (arrays)

        Returns
        -------

        wave_vector: float
            average wave vector from the reference frame
        """
        if len(serie) == 0:
            logger.warning("0 ref image. Use of default k_x = 40.625.")
            return 40.625
        names = serie.get_path_arrays()
        ref = np.zeros((self.ymax - self.ymin, self.xmax - self.xmin))
        ii = 0

        for name in names:
            array = imread(str(Path(self.path_ref) / name))
            frame = array[self.ymin : self.ymax, self.xmin : self.xmax].astype(
                float
            )
            frame = self.frame_normalize(frame)
            ref = ref + frame
            ii += 1
        self.ref = ref / ii
        return self.wave_vector(
            self.ref, self.ymin, self.ymax, self.xmin, self.xmax, self.sur
        )

    def set_gain_filter(self, k_x, l_y, l_x, slicer):
        """compute gain and filter"""
        kx = np.arange(-l_x / 2, l_x / 2) / l_x
        ky = np.arange(-l_y / 2, l_y / 2) / l_y
        kxgrid, kygrid = np.meshgrid(kx, ky)
        X, Y = np.meshgrid(kx * l_x, ky * l_y)
        gain = np.exp(-1.0j * 2 * np.pi * (k_x / l_x * X))
        filt1 = np.fft.fftshift(
            np.exp(-((kxgrid**2 + kygrid**2) / 2 / (k_x / slicer / l_x) ** 2))
            * np.exp(1 - 1 / (1 + ((kxgrid + k_x) ** 2 + kygrid**2) / k_x**2))
        )

        filt2 = np.fft.fftshift(
            -np.exp(
                -(
                    ((kxgrid + (k_x / l_x)) ** 2 + kygrid**2)
                    / 2
                    / (k_x / 10 / l_x) ** 2
                )
            )
            + 1
        )
        filt3 = np.fft.fftshift(
            -np.exp(
                -(
                    ((kxgrid - (k_x / l_x)) ** 2 + kygrid**2)
                    / 2
                    / (k_x / 10 / l_x) ** 2
                )
            )
            + 1
        )
        return gain, filt1 * filt2 * filt3

    def get_borders(self, frame):
        """find the left and right border of the surface in a given frame

        Parameters
        ----------

        frame: int
            frame of the high speed video (np.array)

        Returns
        -------
        xmin: int
            left border of the structure
        xmax: int
            right border of the structure
        """
        frame_thres = 1.0 * (frame > self.thres)
        a = np.argmax(frame_thres, axis=1)
        xmin = np.median(a)
        b = np.argmax(frame_thres[:, ::-1], axis=1)
        b_med = np.median(b)
        xmax = frame.shape[1] - b_med
        return int(xmin), int(xmax)

    def merge_cropped_frame(self, frame, x_min, x_max):
        """puts the actual frame in the reference plate frame to avoid jerks
        and to keep the dimensions

        Parameters
        ----------

        frame: int
            frame of the high speed video (np.array)

        Returns
        -------

        calc_frame: int
            frame of the reference size with embedded smaller structure
        """
        if x_max >= self.xmax:
            x_max = self.xmax - 1
            print("INFO:x_max adjusted")
        if x_min <= self.xmin:
            x_min = self.xmin
            print("INFO:x_min adjusted")
        calc_frame = self.ref
        calc_frame[:, x_min - self.xmin : -(self.xmax - x_max)] = frame[
            self.ymin : self.ymax, x_min:x_max
        ]
        return calc_frame

    def rectify_frame(self, frame, gain, filt):
        """rectify a frame with gain and filt

        Parameters
        ----------

        frame: int
            frame of the high speed video (np.array)
        gain: complex array
            gain for the pattern of the frame

        filt: complex array
            filter for the pattern of the frame        gain:

        Returns
        -------

        rectified frame: int
            array of the rectified frame
        """
        return np.fft.fft2(frame * gain) * filt

    def frame_normalize(self, frame):
        """normalize the frame values by its mean value

        Parameters
        ----------

        frame: int
            frame of the high speed video (np.array)

        Returns
        -------
        normalized_frame: int
            normalized frame of the high speed video (np.array)
        """
        meanx_frame = np.mean(frame, axis=1)
        for y in range(np.shape(frame)[1]):
            frame[:, y] = frame[:, y] / meanx_frame
        normalized_frame = frame - np.mean(frame)
        return normalized_frame

    def process_frame(
        self, frame, ymin, ymax, xmin, xmax, gain, filt, red_factor
    ):
        """process a frame and return phase
        Parameters
        ----------

        frame: int array
            single frame of the high speed video

        xmin: int (default 475)
            xmin to crop the image im[xmin:xmax, ymin:ymax]

        xmax: int (default 640)
            xmax to crop the image im[xmin:xmax, ymin:ymax]

        ymin: int (default 50)
            ymin to crop the image im[xmin:xmax, ymin:ymax]

        ymax: int (default 700)
            ymax to crop the image im[xmin:xmax, ymin:ymax]

        gain: complex array
            gain for the pattern of the frame

        filt: complex array
            filter for the pattern of the frame

        red_factor: int(default 1)
            reduction factor for the frame array to speed up the calc

        Returns
        --------
            a: array containing phase [radians]

        """
        if self.crop_edge is False:
            frame = frame[ymin:ymax, xmin:xmax]
        frame1 = self.frame_normalize(frame).astype(float)
        frame_filtered = self.rectify_frame(frame1, gain, filt)
        inversed_filt = np.fft.ifft2(frame_filtered)
        inversed_filt = inversed_filt[::red_factor, ::red_factor]
        a = np.unwrap(np.angle(inversed_filt), axis=1)  # by lines
        a = np.unwrap(a, axis=0)  # by colums
        return a

    def process_frame_func(self, array_and_path):
        """call process_frame function with surface_tracking parameters

        Parameters
        ----------

        array_and_path : tuple containing array and path

        Returns
        -------

        array_and_path : tuple containing array/phase [radians], frame shape
        and path

        """
        array, path = array_and_path

        x_min = self.xmin
        x_max = self.xmax
        if self.crop_edge:
            x_min, x_max = self.get_borders(array)
            array = self.merge_cropped_frame(array, x_min, x_max)
        shape = (x_min, x_max)
        return (
            self.process_frame(
                array,
                self.ymin,
                self.ymax,
                self.xmin,
                self.xmax,
                self.gain,
                self.filt,
                self.red_factor,
            ),
            shape,
            path,
        )

    def calculheight_func(self, array_and_path):
        """call convphase function with surface_tracking parameters

        Parameters
        ----------

        array_and_path : tuple containing array/phase [radians],
        shape of the frame and path

        Returns
        -------

        height_and_path : tuple containing array/height [m], frame shape
        and path

        """
        array, shape, path = array_and_path
        array_ = []
        for a in array:
            jumps = [
                np.sign(int(a[i + 1] - angle))
                for i, angle in enumerate(a[:-1])
                if a[i + 1] - angle > 0.95 * np.pi
                or a[i + 1] - angle < 0.95 * np.pi
            ]

            mapper = np.zeros(len(jumps))
            smoother = []
            switch = 0

            for i, val in enumerate(mapper):
                if jumps[i] > 0:
                    switch = switch + 1
                if jumps[i] < 0:
                    switch = switch - 1
                if switch >= 1:
                    val = -2 * np.pi * switch
                if switch <= -1:
                    val = 2 * np.pi * switch
                smoother.append(val)
            smoother.insert(0, 0)
            array_smoothed = a + smoother
            array_.append(np.array(array_smoothed))
        array_s = np.array(array_)
        array_s = array_s.T
        for a in array_s:
            jumps = [
                np.sign(int(a[i + 1] - angle))
                for i, angle in enumerate(a[:-1])
                if a[i + 1] - angle > 0.95 * np.pi
                or a[i + 1] - angle < 0.95 * np.pi
            ]

            mapper = np.zeros(len(jumps))
            smoother = []
            switch = 0

            for i, val in enumerate(mapper):
                if jumps[i] > 0:
                    switch = switch + 1
                if jumps[i] < 0:
                    switch = switch - 1
                if switch >= 1:
                    val = -2 * np.pi
                if switch <= -1:
                    val = 2 * np.pi
                smoother.append(val)
            smoother.insert(0, 0)
            array_smoothed = a + smoother
            array_.append(np.array(array_smoothed))
        array_s = array
        return (
            self.convphase(
                array_s,
                self.pix_size,
                self.distance_object,
                self.distance_lens,
                self.wave_proj,
                self.red_factor,
            ),
            shape,
            path,
        )

    def set_borders_zero_func(self, array_and_path):
        """call convphase function with surface_tracking parameters

        Parameters
        ----------

        array_and_path : tuple containing array/phase [radians], shape of
        the frame  and path

        Returns
        -------

        height_and_path : tuple containing array/height [m] and path

        """
        array, shape, path = array_and_path
        (x_min, x_max) = shape
        if x_max >= self.xmax:
            x_max = self.xmax - 1
            logger.warning("x_max adjusted")
        newarray = np.zeros(self.ref.shape)
        newarray[:, self.borders : -self.borders] = resize(
            array[
                :,
                x_min
                + self.borders
                - self.xmin : -(self.xmax - x_max + self.borders),
            ],
            (self.ref.shape[0], self.xmax - 2 * self.borders - self.xmin),
        )
        return (newarray, path)

    def convphase(self, phase, pix_size, dist, dist_p_c, wave_len, red_factor):
        """converts phase array into array of the height [m]

        Parameters
        ----------

        phase : float
            the image phase [radians]

        pix_size  : float
            size of the pixel [m/pixel]

        dist : float
            distance between object and camera [m]

        dist_p_c : float
            distance between projector and camera [m]

        waven_len : float
            wave length of the object [m]

        red_factor: int
            is the reduction factor

        Returns
        -------

        height: float
            array with the height calculated from the phase

        Notes
        -----

        Make sure that the grid is parallel to y

        """

        height = phase * dist / (phase - 2 * np.pi / wave_len * dist_p_c)
        if self.correct_pos is True:
            height = height.astype(float)
            [ld, Ld] = phase.shape
            x = (np.arange(Ld) - Ld / 2) * pix_size * red_factor
            y = (np.arange(ld) - ld / 2) * pix_size * red_factor
            [X, Y] = np.meshgrid(x, y)
            # perform correction
            dX = -X / dist * height
            dY = -Y / dist * height
            dX[1, :] = 0
            dX[-1, :] = 0
            dX[:, 1] = 0
            dX[:-1] = 0
            dY[1, :] = 0
            dY[-1, :] = 0
            dY[:, 1] = 0
            dY[:-1] = 0
            # interploate the values on the new grid
            height = scipy.interpolate.griddata(
                ((X + dX).reshape(ld * Ld), (Y + dY).reshape(ld * Ld)),
                height.reshape(Ld * ld),
                (X, Y),
                method="cubic",
            )
        if self.correct_height is True:
            height = height - self.ref_height
        # height = height-np.mean(np.mean(height))
        return height

    def correctcouple(self, queue_couple):
        """correct phase in order to avoid jump phase"""
        (
            (anglemod, shapemod, path_anglemod),
            (angle, shape, path_angle),
        ) = queue_couple
        fix_y = int(np.fix(self.l_y / 20 / self.red_factor))
        fix_x = int(np.fix(self.l_x / 2 / self.red_factor))
        correct_angle = angle
        jump = angle[fix_y, fix_x] - anglemod[fix_y, fix_x]
        while abs(jump) > np.pi:
            correct_angle = correct_angle - np.sign(jump) * 2 * math.pi
            jump = correct_angle[fix_y, fix_x] - anglemod[fix_y, fix_x]
            print("angle corrected")
        return (correct_angle, shape, path_angle)

    def wave_vector(self, ref, ymin, ymax, xmin, xmax, sur):
        """compute k_x value with mean reference frame

        Parameters
        ----------

        ref: int
            frame with averaged

        Returns
        -------
        wave_vector: float
            average wave vector from the given frame
        """
        Fref = np.fft.fft2(ref, ((ymax - ymin) * sur, (xmax - xmin) * sur))
        kxma = np.arange(-(xmax - xmin) * sur / 2, (xmax - xmin) * sur / 2) / sur
        indc = np.max(np.fft.fftshift(abs(Fref)), axis=0).argmax()
        return abs(kxma[indc])

    def process_ref(self):
        """calculate the reference height from a set of frames of a flat
        plate with zero height
        """
        frame_filtered = self.rectify_frame(self.ref, self.gain, self.filt)
        inversed_filt = np.fft.ifft2(frame_filtered)
        inversed_filt = inversed_filt[:: self.red_factor, :: self.red_factor]
        ref_angle = np.unwrap(np.angle(inversed_filt), axis=1)  # by lines
        ref_angle = np.unwrap(ref_angle, axis=0)  # by columsref
        self.ref_height = np.zeros(ref_angle.shape)
        return (
            self.calculheight_func((ref_angle, ref_angle.shape, self.path_ref))[0]
            - self.offset
        )


if "sphinx" in sys.modules:
    params = WorkSurfaceTracking.create_default_params()
    __doc__ += params._get_formatted_docs()



import os
import sys
from pathlib import Path
from logging import ERROR, INFO
from distutils.util import strtobool

from runpy import run_path

from setuptools import setup, find_packages

from transonic.dist import ParallelBuildExt, get_logger
import setuptools_scm

if sys.version_info[:2] < (3, 7):
    raise RuntimeError("Python version >= 3.7 required.")

if "egg_info" in sys.argv:
    level = ERROR
else:
    level = INFO

logger = get_logger("fluidimage")
logger.setLevel(level)

build_dependencies_backends = {
    "pythran": ["pythran>=0.9.7"],
    "cython": ["cython"],
    "python": [],
    "numba": [],
}

TRANSONIC_BACKEND = os.environ.get("FLUIDIMAGE_TRANSONIC_BACKEND", "pythran")

if "DISABLE_PYTHRAN" in os.environ:
    DISABLE_PYTHRAN = strtobool(os.environ["DISABLE_PYTHRAN"])

    if (
        "FLUIDIMAGE_TRANSONIC_BACKEND" in os.environ
        and DISABLE_PYTHRAN
        and TRANSONIC_BACKEND == "pythran"
    ):
        raise ValueError

    if DISABLE_PYTHRAN:
        TRANSONIC_BACKEND = "python"


if TRANSONIC_BACKEND not in build_dependencies_backends:
    raise ValueError(
        f"FLUIDIMAGE_TRANSONIC_BACKEND={TRANSONIC_BACKEND} "
        f"not in {list(build_dependencies_backends.keys())}"
    )

setup_requires = []
setup_requires.extend(build_dependencies_backends[TRANSONIC_BACKEND])

here = Path(__file__).parent.absolute()

path_image_samples = here / "image_samples"
if path_image_samples.exists():
    with open("fluidimage/_path_image_samples.py", "w") as file:
        file.write(
            "from pathlib import Path\n\n"
            f'path_image_samples = Path(r"{path_image_samples}")\n'
        )

# Get the long description from the relevant file
with open("README.rst") as file:
    long_description = file.read()
lines = long_description.splitlines(True)
for i, line in enumerate(lines):
    if line.endswith(":alt: Code coverage\n"):
        iline_coverage = i
        break

long_description = "".join(lines[iline_coverage + 2 :])

# Get the version from the relevant file
d = run_path("fluidimage/_version.py")
__version__ = d["__version__"]


def write_rev(rev):
    with open("fluidimage/_hg_rev.py", "w") as file:
        file.write(f'hg_rev = "{rev}"\n')


try:
    full_version = setuptools_scm.get_version()
except (LookupError, OSError):
    revision = "?"
else:
    try:
        revision = full_version.split("+")[1]
    except IndexError:
        revision = full_version
    if "." in revision:
        revision = revision.split(".")[0]
write_rev(revision)


def transonize():

    from transonic.dist import make_backend_files

    paths = [
        "fluidimage/calcul/correl.py",
        "fluidimage/topologies/example.py",
        "fluidimage/calcul/interpolate/thin_plate_spline.py",
        "fluidimage/calcul/subpix.py",
    ]
    make_backend_files([here / path for path in paths])


def create_pythran_extensions():
    import numpy as np
    from transonic.dist import init_pythran_extensions

    compile_arch = os.getenv("CARCH", "native")
    extensions = init_pythran_extensions(
        "fluidimage",
        include_dirs=np.get_include(),
        compile_args=("-O3", f"-march={compile_arch}", "-DUSE_XSIMD"),
    )
    return extensions


def create_extensions():
    if "egg_info" in sys.argv or "dist_info" in sys.argv:
        return []

    logger.info("Running fluidimage setup.py on platform " + sys.platform)

    transonize()

    ext_modules = create_pythran_extensions()

    if ext_modules:
        logger.info(
            "The following extensions could be built if necessary:\n"
            + "".join([ext.name + "\n" for ext in ext_modules])
        )

    return ext_modules


setup(
    name="fluidimage",
    version=__version__,
    description=("fluid image processing with Python."),
    long_description=long_description,
    keywords="PIV",
    author="Pierre Augier",
    author_email="pierre.augier@legi.cnrs.fr",
    url="https://foss.heptapod.net/fluiddyn/fluidimage",
    python_requires=">=3.6",
    license="CeCILL",
    classifiers=[
        # How mature is this project? Common values are
        # 3 - Alpha
        # 4 - Beta
        # 5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        # actually CeCILL License (GPL compatible license for French laws)
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(exclude=["doc", "include", "scripts"]),
    scripts=["bin/fluidimviewer", "bin/fluidimlauncher"],
    entry_points={
        "console_scripts": ["fluidimviewer-pg = fluidimage.gui.pg_main:main"]
    },
    setup_requires=setup_requires,
    ext_modules=create_extensions(),
    cmdclass={"build_ext": ParallelBuildExt},
)



"""logging (:mod:`fluidimage.util.log`)
=======================================

.. autofunction:: log_memory_usage

.. autofunction:: reset_logger

.. autofunction:: log_debug

.. autofunction:: log_error

"""

from logging import DEBUG, getLogger

from fluiddyn.util import config_logging as _cl_fluiddyn

from .util import _get_txt_memory_usage, cstring

__all__ = ["logger", "DEBUG", "reset_logger", "log_memory_usage"]

logger = getLogger("fluidimage")


def config_logging(level="info", name="fluidimage", file=None):
    _cl_fluiddyn(level=level, name=name, file=file)


def reset_logger():
    """Remove all handlers (files) linked to the fluidimage logger"""
    for handler in logger.handlers:
        logger.removeHandler(handler)


def log_memory_usage(string="Memory usage", color="OKGREEN", mode="info"):
    """Log the memory usage."""

    if mode == "debug":
        log = logger.debug
    elif mode == "error":
        log = logger.error
    else:
        log = logger.info

    log(_get_txt_memory_usage(string, color))


def log_debug(string):
    """Log in debug mode with WARNING color"""
    logger.debug(cstring(string, color="WARNING"))


def log_error(string):
    """Log in error mode with FAIL color"""
    logger.error(cstring(string, color="FAIL"))



"""Task runner for the developer

Usage
-----

   nox -l

   nox -s <session>

   nox -k <keyword>
or:

   make <session>

execute ``make list-sessions```` or ``nox -l`` for a list of sessions.

"""
import re
import shlex
from pathlib import Path
from functools import partial

import nox

PACKAGE = "snek5000"
CWD = Path.cwd()

if (CWD / "poetry.lock").exists():
    BUILD_SYSTEM = "poetry"
    PACKAGE_SPEC = "pyproject.toml"
else:
    BUILD_SYSTEM = "setuptools"
    PACKAGE_SPEC = "setup.cfg"

TEST_ENV_VARS = {}

EXTRA_REQUIRES = ("main", "doc", "test", "dev")


@nox.session(name="pip-compile", reuse_venv=True)
@nox.parametrize(
    "extra", [nox.param(extra, id=extra) for extra in EXTRA_REQUIRES]
)
def pip_compile(session, extra):
    """Pin dependencies to requirements/*.txt

    How to run all::

        pipx install nox
        nox -s pip-compile

    """
    session.install("pip-tools")
    req = Path("requirements")

    if extra == "main":
        in_extra = ""
        in_file = ""
    else:
        in_extra = f"--extra {extra}"
        in_file = req / "vcs_packages.in"

    out_file = req / f"{extra}.txt"

    session.run(
        *shlex.split(
            "python -m piptools compile --resolver backtracking --quiet "
            f"{in_extra} {in_file} {PACKAGE_SPEC} "
            f"-o {out_file}"
        ),
        *session.posargs,
    )

    session.log(f"Removing absolute paths from {out_file}")
    packages = out_file.read_text()
    rel_path_packages = packages.replace(
        "file://" + str(Path.cwd().resolve()), "."
    )

    if extra == "tests":
        tests_editable = out_file.parent / out_file.name.replace(
            "tests", "tests-editable"
        )
        session.log(f"Copying {out_file} with -e flag in {tests_editable}")
        tests_editable.write_text(rel_path_packages)
        session.log(f"Removing -e flag in {out_file}")
        rel_path_packages = re.sub(r"^-e\ \.", ".", rel_path_packages, flags=re.M)

    session.log(f"Writing {out_file}")
    out_file.write_text(rel_path_packages)


no_venv_session = partial(nox.session, venv_backend="none")
nox.options.sessions = ["tests"]


@nox.session
def tests(session):
    """Execute unit-tests using pytest"""

    session.install("-r", "requirements/test.txt")
    session.run(
        "python",
        "-m",
        "pytest",
        *session.posargs,
        env=TEST_ENV_VARS,
    )


@no_venv_session(name="tests-cov")
def tests_cov(session):
    """Execute unit-tests using pytest+pytest-cov"""
    session.notify(
        "tests",
        [
            "--cov",
            # "--cov-config=setup.cfg",
            "--no-cov-on-fail",
            "--cov-report=term-missing",
            *session.posargs,
        ],
    )


@nox.session(name="coverage-html")
def coverage_html(session, nox=False):
    """Generate coverage report in HTML. Requires `tests-cov` session."""
    report = Path.cwd() / ".coverage" / "html" / "index.html"
    session.install("coverage[toml]")
    session.run("coverage", "html")

    print("Code coverage analysis complete. View detailed report:")
    print(f"file://{report}")



import unittest

from numpy.testing import assert_almost_equal, assert_array_equal

from fluiddyn.io import stdout_redirected
from fluidimage import get_path_image_samples
from fluidimage.calibration.calib_cv import (
    CalibCV,
    SimpleCircleGrid,
    construct_object_points,
)
from fluidimage.util import imread

pathbase = get_path_image_samples() / "TomoPIV" / "calibration"


def construct_image_points(path, origin):
    image = imread(str(path))
    # image = image.astype(np.uint8)
    # Rescale
    scale = image.max()
    image = image * (256 / scale)

    params = SimpleCircleGrid.create_default_params()
    circle_grid = SimpleCircleGrid(params)
    centers = circle_grid.detect_grid(image, origin, nx=7, ny=7, ds=50)

    return centers


class TestCalib(unittest.TestCase):
    """Test fluidimage.calibration.calib_cv module."""

    @classmethod
    def setUpClass(cls):
        path_cache = pathbase / "cam0.h5"
        # with stdout_redirected():
        cls.calib_cache = CalibCV(str(path_cache))

    def test_interpolate(self):
        """Test interpolation."""
        self.calib_cache.get_rotation(2)
        self.calib_cache.get_translate(-2)

    def test_calibrate(self):
        """Tests construct_object_points and CalibCV methods."""
        path_input = pathbase / "cam0" / "0mm_cam0.tif"
        with stdout_redirected():
            calib = CalibCV("fluidimage_test_calib_cv.h5")

        result_cache = self.calib_cache.params

        origin = (250, 250)  # Hardcoded origin
        imgpoints = [construct_image_points(path_input, origin)]
        objpoints = [construct_object_points(7, 7, 0.0, 3.0)]
        zs = [0]
        im_shape = imread(str(path_input)).shape[::-1]
        ret, mtx, dist, rvecs, tvecs = calib.calibrate(
            imgpoints, objpoints, zs, im_shape, origin, debug=True
        )

        assert_array_equal(mtx, result_cache.cam_mtx)
        assert_almost_equal(rvecs[0], result_cache.rotation[2])
        assert_almost_equal(tvecs[0], result_cache.translate[2])


if __name__ == "__main__":
    unittest.main()



"""Executor async/await + multiprocessing (:mod:`fluidimage.executors.exec_async_multiproc`)
============================================================================================

A executor using async for IO and multiprocessing for CPU bounded tasks.

.. autoclass:: ExecutorAsyncMultiproc
   :members:
   :private-members:

"""

import time
from multiprocessing import Event, Pipe, Process

import trio

from fluidimage.util import cstring, log_debug, log_memory_usage, logger

from .exec_async import ExecutorAsync


class ExecutorAsyncMultiproc(ExecutorAsync):
    """Async executor using multiprocessing to launch CPU-bounded tasks"""

    async def async_run_work_cpu(self, work):
        """Is destined to be started with a "trio.start_soon".

        Executes the work on an item (key, obj), and add the result on
        work.output_queue.

        Parameters
        ----------

        work :

          A work from the topology

        """
        self.nb_working_workers_cpu += 1

        try:
            key, obj = work.input_queue.pop_first_item()
        except KeyError:
            self.nb_working_workers_cpu -= 1
            return

        if work.check_exception(key, obj):
            self.nb_working_workers_cpu -= 1
            return

        t_start = time.time()
        log_memory_usage(
            f"{time.time() - self.t_start:.2f} s. Launch work "
            + work.name_no_space
            + f" ({key}). mem usage"
        )

        def exec_work_and_comm(func, obj, child_conn, event):
            # log_debug(f"process ({key}) started")
            event.set()
            # pylint: disable=W0703
            try:
                result = func(obj)
            except Exception as error:
                result = error

            # log_debug(f"in process, send result ({key}): {result}")
            child_conn.send(result)

        parent_conn, child_conn = Pipe()
        event = Event()

        def run_process():

            # we do this complicate thing because there may be a strange bug

            def start_process_and_check(index_attempt):
                process = Process(
                    target=exec_work_and_comm,
                    args=(work.func_or_cls, obj, child_conn, event),
                )
                process.daemon = True
                process.start()
                # check whether the process has really started (possible bug!)
                if not event.wait(1):
                    log_debug(
                        f"problem: process {work.name_no_space} ({key}) "
                        f"has not really started... (attempt {index_attempt})"
                    )
                    process.terminate()
                    return False
                return process

            really_started = False
            for index_attempt in range(10):
                process = start_process_and_check(index_attempt)
                if process:
                    really_started = True
                    break

            if not really_started:
                raise Exception(
                    f"A process {work.name_no_space} ({key}) "
                    "has not started after 10 attempts"
                )

            # todo: use parent_conn.poll to implement a timeout

            # log_debug(f"waiting for result ({key})")
            result = parent_conn.recv()
            # log_debug(f"result ({key}) received")

            process.join(10 * self.sleep_time)
            if process.exitcode != 0:
                logger.info(f"process.exitcode: {process.exitcode}")
                process.terminate()

            return result

        ret = await trio.to_thread.run_sync(run_process)

        if isinstance(ret, Exception):
            self.log_exception(ret, work.name_no_space, key)
            if self.stop_if_error:
                raise ret
        else:
            logger.info(
                f"work {work.name_no_space} ({key}) "
                f"done in {time.time() - t_start:.3f} s"
            )

        if work.output_queue is not None:
            work.output_queue[key] = ret
        self.nb_working_workers_cpu -= 1



"""Piv work and subworks
========================

.. todo::

   - fix using secondary peak in correlation

   - displacement_max

   - displacement_mean


.. autoclass:: BaseWorkPIV
   :members:
   :private-members:

.. autoclass:: FirstWorkPIV
   :members:
   :private-members:

.. autoclass:: WorkPIVFromDisplacement
   :members:
   :private-members:

"""

from copy import deepcopy

import numpy as np

from fluiddyn.util.serieofarrays import SerieOfArraysFromFiles

from ..with_mask import BaseWorkWithMask
from ...calcul.correl import correlation_classes
from ...calcul.errors import PIVError
from ...calcul.interpolate.griddata import griddata
from ...calcul.interpolate.thin_plate_spline_subdom import ThinPlateSplineSubdom
from ...calcul.subpix import SubPix
from ...data_objects.piv import ArrayCouple, HeavyPIVResults


class InterpError(ValueError):
    pass


def _isint(obj):
    return isinstance(obj, (int, np.integer))


class BaseWorkPIV(BaseWorkWithMask):
    """Base class for PIV.

    This class is meant to be subclassed, not instantiated directly.

    """

    @classmethod
    def _complete_params_with_default(cls, params):
        pass

    def _init_shape_crop(self, shape_crop_im0, shape_crop_im1):

        if shape_crop_im1 is None:
            shape_crop_im1 = shape_crop_im0

        if _isint(shape_crop_im0):
            shape_crop_im0 = (shape_crop_im0, shape_crop_im0)
        elif (
            isinstance(shape_crop_im0, (tuple, list)) and len(shape_crop_im0) == 2
        ):
            shape_crop_im0 = tuple(shape_crop_im0)
        else:
            raise NotImplementedError(
                "For now, shape_crop_im0 has to be one or two integer!"
            )

        if _isint(shape_crop_im1):
            shape_crop_im1 = (shape_crop_im1, shape_crop_im1)
        elif (
            isinstance(shape_crop_im1, (tuple, list)) and len(shape_crop_im1) == 2
        ):
            shape_crop_im1 = tuple(shape_crop_im1)
        else:
            raise NotImplementedError(
                "For now, shape_crop_im1 has to be one or two integer!"
            )

        if (
            shape_crop_im1[0] > shape_crop_im0[0]
            or shape_crop_im1[1] > shape_crop_im0[1]
        ):
            raise ValueError(
                "shape_crop_im1 must be inferior or equal to shape_crop_im0"
            )

        self.shape_crop_im0 = tuple(int(n) for n in shape_crop_im0)
        self.shape_crop_im1 = tuple(int(n) for n in shape_crop_im1)

    def _init_correl(self):
        try:
            correl_cls = correlation_classes[self.params.piv0.method_correl]
        except KeyError:
            raise ValueError(
                "params.piv0.method_correl should be in "
                + str(list(correlation_classes.keys()))
            )

        self.correl = correl_cls(
            im0_shape=self.shape_crop_im0,
            im1_shape=self.shape_crop_im1,
            method_subpix=self.params.piv0.method_subpix,
            nsubpix=self.params.piv0.nsubpix,
            displacement_max=self.params.piv0.displacement_max,
            particle_radius=self.params.piv0.particle_radius,
            nb_peaks_to_search=self.params.piv0.nb_peaks_to_search,
        )

    def _prepare_with_image(self, im0=None, imshape=None):
        """Initialize the object with an image."""
        if imshape is None:
            imshape = im0.shape

        self.imshape0 = (len_y, len_x) = imshape
        scim = self.shape_crop_im0

        stepy = scim[0] - int(np.round(self.overlap * scim[0]))
        stepx = scim[1] - int(np.round(self.overlap * scim[1]))
        assert stepy >= 1
        assert stepx >= 1

        ixvec_max = len_x - self._stop_for_crop0[1]
        ixvecs = np.arange(self._start_for_crop0[1], ixvec_max, stepx, dtype=int)

        iyvec_max = len_y - self._stop_for_crop0[0]
        iyvecs = np.arange(self._start_for_crop0[0], iyvec_max, stepy, dtype=int)

        # There are some cases for which it is worth to add the last points...
        if ixvec_max - ixvecs[-1] > stepx // 2.5:
            # print('add another point (x)')
            ixvecs = np.append(ixvecs, ixvec_max)
        if iyvec_max - iyvecs[-1] > stepy // 2.5:
            # print('add another point (y)')
            iyvecs = np.append(iyvecs, iyvec_max)

        self.ixvecs = ixvecs
        self.iyvecs = iyvecs

        ixvecs, iyvecs = np.meshgrid(ixvecs, iyvecs)

        self.ixvecs_grid = ixvecs.flatten()
        self.iyvecs_grid = iyvecs.flatten()

    def calcul(self, couple):
        """Calcul the PIV (one pass) from a couple of images."""
        if isinstance(couple, SerieOfArraysFromFiles):
            couple = ArrayCouple(serie=couple)

        if not isinstance(couple, ArrayCouple):
            raise ValueError

        couple.apply_mask(self.params.mask)

        im0, im1 = couple.get_arrays()
        if not hasattr(self, "ixvecs_grid"):
            self._prepare_with_image(im0)
        (
            deltaxs,
            deltays,
            xs,
            ys,
            correls_max,
            correls,
            errors,
            secondary_peaks,
        ) = self._loop_vectors(im0, im1)

        xs, ys = self._xyoriginalimage_from_xymasked(xs, ys)

        result = HeavyPIVResults(
            deltaxs,
            deltays,
            xs,
            ys,
            errors,
            correls_max=correls_max,
            correls=correls,
            couple=deepcopy(couple),
            params=self.params,
            secondary_peaks=secondary_peaks,
        )

        self._complete_result(result)

        return result

    def _complete_result(self, result):
        result.indices_no_displacement = self.correl.get_indices_no_displacement()
        result.displacement_max = self.correl.displacement_max

    def _pad_images(self, im0, im1):
        """Pad images with zeros.

        .. todo::

           Choose correctly the variable npad.

        """
        npad = self.npad = max(self._start_for_crop0 + self._stop_for_crop0)
        tmp = [(npad, npad), (npad, npad)]
        im0pad = np.pad(im0 - im0.min(), tmp, "constant")
        im1pad = np.pad(im1 - im1.min(), tmp, "constant")
        return im0pad, im1pad

    def _calcul_indices_vec(self, deltaxs_approx=None, deltays_approx=None):
        """Calcul the indices corresponding to the vectors and cropped windows.

        Returns
        -------

        xs : np.array

          x index of the position of the computed vector in the original
          images.

        ys : np.array

          y index of the position of the computed vector in the original
          images.

        ixs0_pad : np.array

          x index of the center of the crop image 0 in the padded image 0.

        iys0_pad : np.array

          y index of the center of the crop image 0 in the padded image 0.

        ixs1_pad : np.array

          x index of the center of the crop image 1 in the padded image 1.

        iys1_pad : np.array

          y index of the center of the crop image 1 in the padded image 1.

        """
        ixs0_pad = self.ixvecs_grid + self.npad
        iys0_pad = self.iyvecs_grid + self.npad
        ixs1_pad = ixs0_pad
        iys1_pad = iys0_pad

        return (
            self.ixvecs_grid,
            self.iyvecs_grid,
            ixs0_pad,
            iys0_pad,
            ixs1_pad,
            iys1_pad,
        )

    def _loop_vectors(self, im0, im1, deltaxs_approx=None, deltays_approx=None):
        """Loop over the vectors to compute them."""

        im0pad, im1pad = self._pad_images(im0, im1)

        xs, ys, ixs0_pad, iys0_pad, ixs1_pad, iys1_pad = self._calcul_indices_vec(
            deltaxs_approx=deltaxs_approx, deltays_approx=deltays_approx
        )

        nb_vec = len(xs)

        correls = [None] * nb_vec
        errors = {}
        deltaxs = np.empty(xs.shape, dtype="float32")
        deltays = np.empty_like(deltaxs)
        correls_max = np.empty_like(deltaxs)
        secondary_peaks = [None] * nb_vec

        has_to_apply_subpix = self.index_pass == self.params.multipass.number - 1

        for ivec in range(nb_vec):

            ixvec0 = ixs0_pad[ivec]
            iyvec0 = iys0_pad[ivec]
            ixvec1 = ixs1_pad[ivec]
            iyvec1 = iys1_pad[ivec]

            im0crop = self._crop_im0(ixvec0, iyvec0, im0pad)
            im1crop = self._crop_im1(ixvec1, iyvec1, im1pad)

            if (
                im0crop.shape != self.shape_crop_im0
                or im1crop.shape != self.shape_crop_im1
            ):

                print(
                    "Warning: Bad im_crop shape.",
                    ixvec0,
                    iyvec0,
                    ixvec1,
                    iyvec1,
                    im0crop.shape,
                    self.shape_crop_im0,
                    im1crop.shape,
                    self.shape_crop_im1,
                )

                deltaxs[ivec] = np.nan
                deltays[ivec] = np.nan
                correls_max[ivec] = np.nan
                errors[ivec] = "Bad im_crop shape."
                continue

            if np.isnan(im0crop).any() or np.isnan(im1crop).any():
                deltaxs[ivec] = np.nan
                deltays[ivec] = np.nan
                correls_max[ivec] = np.nan
                errors[ivec] = "Nan(s) in cropped images."
                continue

            # compute and store correlation map
            correl, norm = self.correl(im0crop, im1crop)
            if (
                self.index_pass == 0
                and self.params.piv0.coef_correl_no_displ is not None
            ):
                correl[
                    self.correl.get_indices_no_displacement()
                ] *= self.params.piv0.coef_correl_no_displ

            correls[ivec] = correl

            # compute displacements corresponding to peaks
            try:
                (
                    deltax,
                    deltay,
                    correl_max,
                    other_peaks,
                ) = self.correl.compute_displacements_from_correl(
                    correl, norm=norm
                )
            except PIVError as e:
                errors[ivec] = e.explanation
                deltaxs[ivec], deltays[ivec], correls_max[ivec] = e.results
                continue

            # increase precision on the displacement
            if has_to_apply_subpix:
                try:
                    deltax, deltay = self.correl.apply_subpix(
                        deltax, deltay, correl
                    )
                except PIVError as e:
                    errors[ivec] = e.explanation

            deltaxs[ivec] = deltax
            deltays[ivec] = deltay
            correls_max[ivec] = correl_max

            secondary_peaks[ivec] = other_peaks

        if deltaxs_approx is not None:
            deltaxs += deltaxs_approx
            deltays += deltays_approx

        return (
            deltaxs,
            deltays,
            xs,
            ys,
            correls_max,
            correls,
            errors,
            secondary_peaks,
        )

    def _init_crop(self):
        """Initialize the cropping of the images."""

        scim0 = self.shape_crop_im0
        scim1 = self.shape_crop_im1

        self._start_for_crop0 = (scim0[0] // 2, scim0[1] // 2)
        _stop_for_crop0 = [scim0[0] // 2, scim0[1] // 2]

        if scim0[0] % 2 == 1:
            _stop_for_crop0[0] += 1

        if scim0[1] % 2 == 1:
            _stop_for_crop0[1] += 1

        self._stop_for_crop0 = tuple(_stop_for_crop0)

        self._start_for_crop1 = (scim1[0] // 2, scim1[1] // 2)
        _stop_for_crop1 = [scim1[0] // 2, scim1[1] // 2]

        if scim1[0] % 2 == 1:
            _stop_for_crop1[0] += 1

        if scim1[1] % 2 == 1:
            _stop_for_crop1[1] += 1

        self._stop_for_crop1 = tuple(_stop_for_crop1)

    def _crop_im0(self, ixvec, iyvec, im):
        """Crop image 0."""
        subim = im[
            iyvec - self._start_for_crop0[0] : iyvec + self._stop_for_crop0[0],
            ixvec - self._start_for_crop0[1] : ixvec + self._stop_for_crop0[1],
        ]
        subim = np.array(subim, dtype=np.float32)
        return subim - subim.mean()

    def _crop_im1(self, ixvec, iyvec, im):
        """Crop image 1."""
        subim = im[
            iyvec - self._start_for_crop1[0] : iyvec + self._stop_for_crop1[0],
            ixvec - self._start_for_crop1[1] : ixvec + self._stop_for_crop1[1],
        ]
        subim = np.array(subim, dtype=np.float32)
        return subim - subim.mean()

    def apply_interp(self, piv_results, last=False):
        """Interpolate a PIV result object on the grid of the PIV work.

        Parameters
        ----------

        piv_results : :class;`HeavyPIVResults`

            The interpolated field are added to this object.

        last : bool (False)

            Last pass or not.

        Notes
        -----

        Depending on the value of `params.multipass.use_tps`, the interpolation
        is done with an interative method based on the Thin Plate Spline method
        (:class:`fluidimage.calcul.interpolate.thin_plate_spline_subdom.ThinPlateSplineSubdom`)
        or done with a simple griddata method (much faster).

        """

        if not last and not hasattr(piv_results, "ixvecs_approx"):
            (
                piv_results.ixvecs_approx,
                piv_results.iyvecs_approx,
            ) = self._xyoriginalimage_from_xymasked(
                self.ixvecs_grid, self.iyvecs_grid
            )

        if last and not hasattr(piv_results, "ixvecs_final"):
            (
                piv_results.ixvecs_final,
                piv_results.iyvecs_final,
            ) = self._xyoriginalimage_from_xymasked(
                self.ixvecs_grid, self.iyvecs_grid
            )

        # for the interpolation
        selection = ~(
            np.isnan(piv_results.deltaxs) | np.isnan(piv_results.deltays)
        )

        if not any(selection):
            raise InterpError("Only nan.")

        xs = piv_results.xs[selection]
        ys = piv_results.ys[selection]

        xs, ys = self._xymasked_from_xyoriginalimage(xs, ys)

        centers = np.vstack([ys, xs])

        deltaxs = piv_results.deltaxs[selection]
        deltays = piv_results.deltays[selection]

        if (
            self.params.multipass.use_tps is True
            or self.params.multipass.use_tps == "last"
            and last
        ):
            print(f"TPS interpolation ({piv_results.couple.name}).")
            # compute TPS coef
            smoothing_coef = self.params.multipass.smoothing_coef
            subdom_size = self.params.multipass.subdom_size
            threshold = self.params.multipass.threshold_tps

            tps = ThinPlateSplineSubdom(
                centers,
                subdom_size,
                smoothing_coef,
                threshold=threshold,
                percent_buffer_area=0.25,
            )
            try:
                deltaxs_smooth, deltaxs_tps = tps.compute_tps_coeff_subdom(
                    deltaxs
                )
                deltays_smooth, deltays_tps = tps.compute_tps_coeff_subdom(
                    deltays
                )
            except np.linalg.LinAlgError:
                print("compute delta_approx with griddata (in tps)")
                deltaxs_approx = griddata(
                    centers[::-1], deltaxs, (self.ixvecs, self.iyvecs)
                )
                deltays_approx = griddata(
                    centers[::-1], deltays, (self.ixvecs, self.iyvecs)
                )
            else:
                piv_results.deltaxs_smooth = deltaxs_smooth
                piv_results.deltaxs_tps = deltaxs_tps
                piv_results.deltays_smooth = deltays_smooth
                piv_results.deltays_tps = deltays_tps

                new_positions = np.vstack([self.iyvecs_grid, self.ixvecs_grid])

                tps.init_with_new_positions(new_positions)

                deltaxs_approx = tps.compute_eval(deltaxs_tps)
                deltays_approx = tps.compute_eval(deltays_tps)
        else:
            deltaxs_approx = griddata(
                centers[::-1], deltaxs, (self.ixvecs, self.iyvecs)
            )
            deltays_approx = griddata(
                centers[::-1], deltays, (self.ixvecs, self.iyvecs)
            )

        if last:
            piv_results.deltaxs_final = deltaxs_approx
            piv_results.deltays_final = deltays_approx
        else:
            piv_results.deltaxs_approx = deltaxs_approx
            piv_results.deltays_approx = deltays_approx


class FirstWorkPIV(BaseWorkPIV):
    """First PIV pass (without input displacements).

    Parameters
    ----------

    params : ParamContainer

      ParamContainer object produced by the function
      :func:`fluidimage.works.piv.multipass.WorkPIV.create_default_params`.

    """

    index_pass = 0

    @classmethod
    def _complete_params_with_default(cls, params):
        params._set_child(
            "piv0",
            attribs={
                "shape_crop_im0": 48,
                "shape_crop_im1": None,
                "displacement_max": None,
                "displacement_mean": None,  # NotImplemented
                "method_correl": "fftw",
                "method_subpix": "2d_gaussian2",
                "nsubpix": None,
                "coef_correl_no_displ": None,
                "nb_peaks_to_search": 1,
                "particle_radius": 3,
            },
        )

        params.piv0._set_doc(
            """Parameters describing one PIV step.

shape_crop_im0 : int (48)
    Shape of the cropped images 0 from which are computed the correlation.
shape_crop_im1 : int or None
    Shape of the cropped images 0 (has to be None for correl based on fft).

displacement_max : None
    Displacement maximum used in correlation classes. The exact effect depends
    on the correlation method. For fft based correlation, it can also be of the
    form '50%' and then the maximum displacement is computed for each pass as a
    pourcentage of max(shape_crop_im0).

displacement_mean : None
    Displacement averaged over space (NotImplemented).

method_correl : str, default 'fftw'

    Can be in """
            + str(list(correlation_classes.keys()))
            + """

method_subpix : str, default '2d_gaussian2'

    Can be in """
            + str(SubPix.methods)
            + """

nsubpix : None
    Integer used in the subpix finder. It is related to the typical size of the
    particles. It has to be increased in case of peak locking (plot the
    histograms of the displacements).

coef_correl_no_displ : None, number
    If this coefficient is not None, the correlation of the point corresponding
    to no displacement is multiplied by this coefficient (for the first pass).

nb_peaks_to_search : 1, int
    Number of peaks to search. Secondary peaks can be used during the fix step.

particle_radius : 3, int
    Typical radius of a particle (or more preciselly of a correlation
    peak). Used only if `nb_peaks_to_search` is larger than one.

"""
        )

        params.piv0._set_child(
            "grid", attribs={"overlap": 0.5, "from": "overlap"}
        )

        params.piv0.grid._set_doc(
            """
Parameters describing the grid.

overlap : float (0.5)
    Number smaller than 1 defining the overlap between interrogation windows.

from : str {'overlap'}
    Keyword for the method from which is computed the grid.
"""
        )

        cls._complete_params_with_default_mask(params)

    def __init__(self, params):

        self.params = params

        self.overlap = params.piv0.grid.overlap
        if self.overlap >= 1:
            raise ValueError("params.piv0.grid.overlap has to be smaller than 1")

        self._init_shape_crop(
            params.piv0.shape_crop_im0, params.piv0.shape_crop_im1
        )
        self._init_crop()
        self._init_correl()


class WorkPIVFromDisplacement(BaseWorkPIV):
    """Work PIV working from already computed displacement (for multipass).

    Parameters
    ----------

    params : ParamContainer

      ParamContainer object produced by the function
      :func:`fluidimage.works.piv.multipass.WorkPIV.create_default_params`.

    index_pass : int, 1

    shape_crop_im0 : int or tuple, optional

    shape_crop_im1 : int or tuple, optional

    Notes
    -----

    Steps for the PIV computation:

    - prepare the work PIV with an image if not already done (grid),

    - apply interpolation (TPS or griddata) to compute a estimation of the
      displacements,

    - loop over vectors to compute displacements.

    """

    def __init__(
        self, params, index_pass=1, shape_crop_im0=None, shape_crop_im1=None
    ):

        self.params = params
        self.index_pass = index_pass

        self.overlap = params.piv0.grid.overlap

        if shape_crop_im0 is None:
            shape_crop_im0 = params.piv0.shape_crop_im0
        if shape_crop_im1 is None:
            shape_crop_im1 = params.piv0.shape_crop_im1

        self._init_shape_crop(shape_crop_im0, shape_crop_im1)
        self._init_crop()
        self._init_correl()

    def calcul(self, piv_results):
        """Calcul the PIV (one pass) from a couple of images and displacement.

        .. todo::

           Use the derivatives of the velocity to distort the image 1.

        """
        if not isinstance(piv_results, HeavyPIVResults):
            raise ValueError

        couple = piv_results.couple

        im0, im1 = couple.get_arrays()
        if not hasattr(self, "ixvecs_grid"):
            self._prepare_with_image(im0)

        self.apply_interp(piv_results)

        deltaxs_approx = piv_results.deltaxs_approx
        deltays_approx = piv_results.deltays_approx

        deltaxs_approx = np.round(deltaxs_approx).astype("int32")
        deltays_approx = np.round(deltays_approx).astype("int32")

        (
            deltaxs,
            deltays,
            xs,
            ys,
            correls_max,
            correls,
            errors,
            secondary_peaks,
        ) = self._loop_vectors(
            im0, im1, deltaxs_approx=deltaxs_approx, deltays_approx=deltays_approx
        )

        xs, ys = self._xyoriginalimage_from_xymasked(xs, ys)

        result = HeavyPIVResults(
            deltaxs,
            deltays,
            xs,
            ys,
            errors,
            correls_max=correls_max,
            correls=correls,
            couple=deepcopy(couple),
            params=self.params,
            secondary_peaks=secondary_peaks,
        )

        self._complete_result(result)
        result.deltaxs_approx0 = deltaxs_approx
        result.deltays_approx0 = deltays_approx

        return result

    def _calcul_indices_vec(self, deltaxs_approx=None, deltays_approx=None):
        """Calcul the indices corresponding to the vectors and cropped windows.

        Returns
        -------

        xs : np.array

          x index of the position of the computed vector in the original
          images.

        ys : np.array

          y index of the position of the computed vector in the original
          images.

        ixs0_pad : np.array

          x index of the center of the crop image 0 in the padded image 0.

        iys0_pad : np.array

          y index of the center of the crop image 0 in the padded image 0.

        ixs1_pad : np.array

          x index of the center of the crop image 1 in the padded image 1.

        iys1_pad : np.array

          y index of the center of the crop image 1 in the padded image 1.

        """
        ixs0 = self.ixvecs_grid - deltaxs_approx // 2
        iys0 = self.iyvecs_grid - deltays_approx // 2
        ixs1 = ixs0 + deltaxs_approx
        iys1 = iys0 + deltays_approx

        # if a point is outside an image => shift of subimages used
        # for correlation
        ind_outside = np.argwhere(
            (ixs0 > self.imshape0[1])
            | (ixs0 < 0)
            | (ixs1 > self.imshape0[1])
            | (ixs1 < 0)
            | (iys0 > self.imshape0[0])
            | (iys0 < 0)
            | (iys1 > self.imshape0[0])
            | (iys1 < 0)
        )

        for ind in ind_outside:
            if (
                (ixs1[ind] > self.imshape0[1])
                or (iys1[ind] > self.imshape0[0])
                or (ixs1[ind] < 0)
                or (iys1[ind] < 0)
            ):
                ixs0[ind] = self.ixvecs_grid[ind] - deltaxs_approx[ind]
                iys0[ind] = self.iyvecs_grid[ind] - deltays_approx[ind]
                ixs1[ind] = self.ixvecs_grid[ind]
                iys1[ind] = self.iyvecs_grid[ind]
            elif (
                (ixs0[ind] > self.imshape0[1])
                or (iys0[ind] > self.imshape0[0])
                or (ixs0[ind] < 0)
                or (iys0[ind] < 0)
            ):
                ixs0[ind] = self.ixvecs_grid[ind]
                iys0[ind] = self.iyvecs_grid[ind]
                ixs1[ind] = self.ixvecs_grid[ind] + deltaxs_approx[ind]
                iys1[ind] = self.iyvecs_grid[ind] + deltays_approx[ind]

        xs = (ixs0 + ixs1) / 2.0
        ys = (iys0 + iys1) / 2.0

        ixs0_pad = ixs0 + self.npad
        iys0_pad = iys0 + self.npad
        ixs1_pad = ixs1 + self.npad
        iys1_pad = iys1 + self.npad

        return xs, ys, ixs0_pad, iys0_pad, ixs1_pad, iys1_pad



"""Utilities for calibration (:mod:`fluidimage.calibration.util`)
=================================================================

"""

from math import cos, sin

import numpy as np

from fluiddyn.util.paramcontainer import ParamContainer, tidy_container


def get_number_from_string(string):

    return [float(word) for word in string.split() if word]


def get_plane_equation(z0, alpha, beta):
    """Return coefficients a, b, c, d of the equation of a plane

    The plane is defined with ax + by + cz + d = 0

    Parameters
    ----------

    z0 : number

    alpha : number

      The angle in radian around x axis

    beta : number

      The angle in radian around y axis

    Notes
    -----

    Works only when 0 or 1 angle ~= 0
    """
    if alpha != 0 and beta != 0:
        raise ValueError("Works only when 0 or 1 angle != 0")

    a = sin(beta)
    b = -sin(alpha) * cos(beta)
    c = cos(alpha) * cos(beta)
    d = -c * z0
    return a, b, c, d


def get_base_from_normal_vector(nx, ny, nz):
    """matrix of base change from a given plane to the fixed plane
    n has to be approximately vertical: i.e. nz approx. 1
    """

    ez = np.array([nx, ny, nz])
    ez = ez / np.linalg.norm(ez)

    ex1, ex2 = 1, 0
    ex3 = -(ex1 * nx + ex2 * ny) / nz
    ex = np.array([ex1, ex2, ex3])
    ex = ex / np.linalg.norm(ex)

    ey = np.cross(ez, ex)
    A = np.vstack([ex, ey, ez]).transpose()
    return A, np.linalg.inv(A)


def make_params_calibration(path_file):
    """Make a tidy parameter container from a UVmat file."""
    params = ParamContainer(tag="calib")

    calib_uvmat = ParamContainer(path_file=str(path_file))
    tidy_container(calib_uvmat)

    calib_uvmat = calib_uvmat["geometry_calib"]

    f = [float(n) for n in np.array(get_number_from_string(calib_uvmat.fx_fy))]
    C = np.array(get_number_from_string(calib_uvmat.cx__cy))
    kc = np.array(calib_uvmat.kc)
    T = np.array(get_number_from_string(calib_uvmat.tx__ty__tz))

    R = []

    for i in range(3):
        R = np.hstack(
            [R, get_number_from_string(calib_uvmat["r_{}".format(i + 1)])]
        )

    if "omc" in calib_uvmat.__dict__:
        omc = np.array(get_number_from_string(calib_uvmat["omc"]))
    else:
        omc = []
        for i in range(3):
            omc = np.hstack([omc, calib_uvmat["omc_{}".format(i + 1)]])

    params._set_attribs({"f": f, "C": C, "kc": kc, "T": T, "R": R, "omc": omc})

    if calib_uvmat.nb_slice is not None:

        nb_slice = np.array(calib_uvmat["nb_slice"])
        zslice_coord = np.zeros([nb_slice, 3])

        if calib_uvmat.nb_slice == 1:
            zslice_coord[:] = get_number_from_string(calib_uvmat["slice_coord"])
            if (
                hasattr(calib_uvmat, "slice_angle")
                and calib_uvmat["slice_angle"] is not None
            ):
                slice_angle = np.zeros([nb_slice, 3])
                slice_angle[:] = get_number_from_string(
                    calib_uvmat["slice_angle"]
                )
            else:
                slice_angle = [0, 0, 0]
        else:
            for i in range(nb_slice):
                zslice_coord[i][:] = get_number_from_string(
                    calib_uvmat["slice_coord_{}".format(i + 1)]
                )

            if (
                hasattr(calib_uvmat, "slice_angle_1")
                and calib_uvmat["slice_angle_1"] is not None
            ):
                slice_angle = np.zeros([nb_slice, 3])
                for i in range(nb_slice):
                    slice_angle[i][:] = get_number_from_string(
                        calib_uvmat["slice_angle_{}".format(i + 1)]
                    )
            else:
                slice_angle = [0, 0, 0]

        params._set_child(
            "slices",
            attribs={
                "nb_slice": nb_slice,
                "zslice_coord": zslice_coord,
                "slice_angle": slice_angle,
            },
        )

    if hasattr(calib_uvmat, "refraction_index"):
        params._set_attrib("refraction_index", calib_uvmat.refraction_index)

    if hasattr(calib_uvmat, "interface_coord"):
        params._set_attrib(
            "interface_coord",
            get_number_from_string(calib_uvmat["interface_coord"]),
        )

    return params



import unittest
from pathlib import Path
from shutil import rmtree

from fluidimage import get_path_image_samples
from fluidimage.topologies.bos import TopologyBOS


class TestBOSNew(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path_input_files = get_path_image_samples() / "Karman/Images"
        cls.postfix = "test_bos_new"

    @classmethod
    def tearDownClass(cls):
        path = cls.path_input_files
        path_out = Path(str(path) + "." + cls.postfix)
        if path_out.exists():
            rmtree(path_out)

    def test_bos_new_multiproc(self):
        params = TopologyBOS.create_default_params()

        params.images.path = str(self.path_input_files)
        params.images.str_slice = "1:3"

        params.piv0.shape_crop_im0 = 32
        params.multipass.number = 2
        params.multipass.use_tps = False

        params.mask.strcrop = ":, 50:500"

        # temporary, avoid a bug on Windows
        params.piv0.method_correl = "pythran"
        params.piv0.shape_crop_im0 = 16

        # compute only few vectors
        params.piv0.grid.overlap = -8

        params.saving.how = "recompute"
        params.saving.postfix = self.postfix

        topology = TopologyBOS(params, logging_level="info")
        topology.compute("exec_async", stop_if_error=True)

        # remove one file
        path_files = list(Path(topology.path_dir_result).glob("bos*"))
        path_files[0].unlink()

        params.saving.how = "complete"
        topology = TopologyBOS(params, logging_level="info")
        topology.compute()


if __name__ == "__main__":
    unittest.main()



import numpy as np

from thin_plate_spline import (
    compute_tps_matrix,
    compute_tps_matrix_numpy,
    compute_tps_matrix_pythran,
)

n0 = 259

n1 = 1000

x = np.random.rand(n0)
y = np.random.rand(n0)

centers = np.vstack([x, y])

x = np.random.rand(n1)
y = np.random.rand(n1)

new_positions = np.vstack([x, y])


"""

%timeit compute_tps_matrix(new_positions, centers)
%timeit compute_tps_matrix_numpy(new_positions, centers)

%timeit compute_tps_matrix_pythran(new_positions, centers)

"""



"""Works - processing
=====================

This subpackage defines some works. A work does a processing. It has
initialization parameters and after initialization is able to produce an output
object from an input object. It can also take more than one input objects
and/or return more than one output objects.

A work is made of one or more work units. In particular, it could be useful to
define input/output and computational works.

.. autosummary::
   :toctree:

   piv
   preproc
   surface_tracking
   optical_flow

"""
from .. import imread


class BaseWork:
    def __init__(self, params=None):
        self.params = params


def load_image(path):
    im = imread(path)
    return im



import unittest

import h5py
import matplotlib.pyplot as plt
import numpy as np

from fluidimage import get_path_image_samples
from fluidimage.calibration import Calibration

plt.show = lambda: 0


pathbase = get_path_image_samples() / "Milestone"


def get_piv_field(path):

    try:
        with h5py.File(path, "r") as file:
            keyspiv = [key for key in file.keys() if key.startswith("piv")]
            keyspiv.sort()
            key = keyspiv[-1]
            X = file[key]["xs"][...]
            Y = file[key]["ys"][...]
            dx = file[key]["deltaxs_final"][...]
            dy = file[key]["deltays_final"][...]
    except Exception:
        print(path)
        raise

    return X, Y, dx, dy


class TestCalib(unittest.TestCase):
    """Test fluidimage.calibration DirectStereoReconstruction, CalibDirect."""

    def test(self):
        path_calib = pathbase / "PCO_top.xml"
        path_v = pathbase / "piv_0000a-b.h5"
        nbypix = 2160

        calib = Calibration(path_calib)

        X, Y, dx, dy = get_piv_field(path_v)

        Xphys, Yphys, Zphys, dxphys, dyphys, dzphys = calib.pix2phys_UV(
            X, Y, dx, dy, index_level=0, nbypix=nbypix
        )
        X, Y = calib.phys2pix(Xphys, Yphys, np.nanmean(Zphys))


if __name__ == "__main__":
    unittest.main()



"""Utilities
============

.. autosummary::
   :toctree:

   util
   log

"""

from fluiddyn.io.image import imsave_h5

from .log import (
    DEBUG,
    config_logging,
    log_debug,
    log_error,
    log_memory_usage,
    logger,
    reset_logger,
)
from .util import cstring, imread, imsave, print_memory_usage, str_short

__all__ = [
    "imread",
    "imsave",
    "imsave_h5",
    "print_memory_usage",
    "logger",
    "reset_logger",
    "log_memory_usage",
    "cstring",
    "str_short",
    "DEBUG",
    "log_debug",
    "log_error",
    "config_logging",
]



import unittest

from fluidimage import SerieOfArraysFromFiles, get_path_image_samples

from .image2image import Im2ImExample, apply_im2im_filter

path_src = get_path_image_samples() / "Karman/Images"


class TestTopoExample(unittest.TestCase):
    def test_apply_im2im_filter(self):

        serie = SerieOfArraysFromFiles(path_src, "2:")
        result = apply_im2im_filter(serie)
        assert len(result) == 3
        result = apply_im2im_filter(serie, Im2ImExample, args_init=(1, 2))
        assert len(result) == 3



"""Direct calibration (:mod:`fluidimage.calibration.calib_direct`)
==================================================================

.. autoclass:: CalibDirect
   :members:
   :private-members:

.. autoclass:: DirectStereoReconstruction
   :members:
   :private-members:

"""
import glob
import warnings
from math import sqrt

import matplotlib.pyplot as plt
import numpy as np
import pylab
from scipy.interpolate import (
    LinearNDInterpolator,
    RegularGridInterpolator,
    griddata,
)

from fluiddyn.util.paramcontainer import ParamContainer, tidy_container

from .util import get_base_from_normal_vector, get_number_from_string

# from scipy.interpolate import CloughTocher2DInterpolator


try:
    from mpl_toolkits.mplot3d import Axes3D
except ImportError:
    pass


class Interpolent:
    pass


def get_points(points_file):
    """Get grid points extracted from calibration images"""
    imgpts = ParamContainer(path_file=points_file)
    tidy_container(imgpts)
    imgpts = imgpts.geometry_calib.source_calib.__dict__
    points = [x for x in imgpts.keys() if "point_" in x or "Point" in x]
    points.sort()

    # [X, Y, Z, x, y]
    results = [[], [], [], [], []]
    # X, Y and Z are coordinate in real space (physical unit)
    # x and y are the indices of the pixel in the image
    for point in points:
        numbers = get_number_from_string(imgpts[point])

        for i, result in enumerate(results):
            result.append(numbers[i])

    results = [np.array(result) for result in results]

    # particularity for Z
    results[2] = results[2][0]

    # particularity for X, Y, Z
    for i in range(3):
        results[i] /= 100.0

    return results


class CalibDirect:
    """Class for direct Calibration
    This calibration determine the equations of optical paths for "each" pixels

    Parameters
    ----------

    glob_str_xml: None
      Path for grid points extracted from calibration images
      This files are given by UVMAT.
      example: 'Images/img*.xml'

    nb_pixels: (None, None)
      Number of pixels in the images

    pth_file: None
      Path of calibration file

    """

    def __init__(self, glob_str_xml=None, shape_img=(None, None), path_file=None):

        if path_file:
            self.load(path_file)
        else:
            self.paths_xml = glob.glob(str(glob_str_xml))
            if len(self.paths_xml) == 0:
                raise ValueError(
                    "No xml file found. \n"
                    'glob_str_xml = "{}"'.format(glob_str_xml)
                )

            self.nb_pixels_x = shape_img[1]
            self.nb_pixels_y = shape_img[0]

    def compute_interpolents(self, interpolator=LinearNDInterpolator):
        """Compute interpolents

        Create an object Interpolent (self.interp_levels) containing
        interpolents to switch from indices of pixels to physical coordinates.

        - indices_pixel2xphys
        - indices_pixel2yphys
        - phys2index_x_pixel
        - phys2index_y_pixel
        - zphys

        There are as many interpolents of each sort as numbers of planes (z
        values).

        """
        interp = Interpolent()
        interp.indices_pixel2xphys = []
        interp.indices_pixel2yphys = []
        interp.phys2index_x_pixel = []
        interp.phys2index_y_pixel = []
        interp.Z = []

        for i, path in enumerate(self.paths_xml):
            # naming convention:
            # X, Y and Z are coordinate in real space (physical unit)
            # x and y are the indices of the pixel in the image
            X, Y, Z, x, y = get_points(path)
            interp.Z.append(Z)

            interp.indices_pixel2xphys.append(interpolator((x, y), X))
            interp.indices_pixel2yphys.append(interpolator((x, y), Y))
            interp.phys2index_x_pixel.append(interpolator((X, Y), x))
            interp.phys2index_y_pixel.append(interpolator((X, Y), y))

        self.interp_levels = interp

    def compute_interpolents_pixel2line(self, nb_lines_x, nb_lines_y, test=False):
        """Compute interpolents for parameters for each optical path.

        The number of optical paths is given by nb_lines_x * nb_lines_y.

        Optical paths are defined with a point x0, y0, z0 and a vector dx, dy,
        dz.

        """

        xtmp = np.unique(np.floor(np.linspace(0, self.nb_pixels_x, nb_lines_x)))
        ytmp = np.unique(np.floor(np.linspace(0, self.nb_pixels_y, nb_lines_y)))

        x, y = np.meshgrid(xtmp, ytmp)
        x = x.transpose()
        y = y.transpose()
        V = np.zeros((x.shape[0], x.shape[1], 6))

        xtrue = []
        ytrue = []
        vtrue = []
        xfalse = []
        yfalse = []
        indi = []
        indj = []

        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                tmp = self.pixel2line(x[i, j], y[i, j])
                if np.isnan(tmp[0]):
                    xfalse.append(x[i, j])
                    yfalse.append(y[i, j])
                    indi.append(i)
                    indj.append(j)
                else:
                    V[i, j, :] = tmp
                    xtrue.append(x[i, j])
                    ytrue.append(y[i, j])
                    vtrue.append(tmp)

        if test:
            titles = ["X0", "Y0", "Z0", "dx", "dy", "dz"]
            for j in range(6):
                pylab.figure()
                pylab.pcolor(x, y, V[:, :, j], shading="nearest")
                pylab.title(titles[j])
                pylab.xlabel("x (pix)")
                pylab.ylabel("y (pix)")
                pylab.colorbar()

            pylab.figure()
            pylab.pcolor(
                x,
                y,
                np.sqrt(V[:, :, 3] ** 2 + V[:, :, 4] ** 2 + V[:, :, 5] ** 2),
                shading="nearest",
            )
            pylab.title("norm(d)")
            pylab.xlabel("x (pix)")
            pylab.ylabel("y (pix)")
            pylab.colorbar()
            plt.show()

        vtrue = np.array(vtrue)
        for j in range(6):
            V[indi, indj, j] = griddata(
                (xtrue, ytrue), vtrue[:, j], (xfalse, yfalse)
            )

        interp = []
        for i in range(6):
            interp.append(RegularGridInterpolator((xtmp, ytmp), V[:, :, i]))

        self.interp_lines = interp

    def pixel2line(self, indx, indy):
        """Compute parameters of the optical path for a pixel

        An optical path is defined with a point x0, y0, z0 and a vector dx, dy,
        dz.

        """
        interp = self.interp_levels
        X = []
        Y = []
        Z = interp.Z
        for i in range(len(Z)):
            X.append((interp.indices_pixel2xphys[i]((indx, indy))))
            Y.append((interp.indices_pixel2yphys[i]((indx, indy))))
        X = np.array(X)
        Y = np.array(Y)
        XYZ = np.vstack([X, Y, Z]).transpose()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            XYZ0 = np.nanmean(XYZ, 0)
        XYZ -= XYZ0
        ind = ~np.isnan(X + Y)
        XYZ = XYZ[ind, :]
        if XYZ.shape[0] > 1:
            arg = np.argsort(XYZ[:, 2])
            XYZ = XYZ[arg, :]
            u, s, v = np.linalg.svd(XYZ, full_matrices=True, compute_uv=1)
            direction = np.cross(v[-1, :], v[-2, :])
            if direction[2] < 0:
                direction = -direction
            return np.hstack([XYZ0, direction])

        else:
            return np.hstack([np.nan] * 6)

    def save(self, pth_file):
        """Save calibration"""
        np.save(
            pth_file,
            [
                self.interp_lines,
                self.paths_xml,
                self.nb_pixels_x,
                self.nb_pixels_y,
            ],
        )

    def load(self, pth_file):
        """Load calibration"""
        tmp = np.load(pth_file, allow_pickle=True)
        self.interp_lines = tmp[0]
        self.paths_xml = tmp[1]
        self.nb_pixels_x = tmp[2]
        self.nb_pixels_y = tmp[3]

    def intersect_with_plane(self, indx, indy, a, b, c, d):
        """Find intersection with the line associated to the pixel  indx, indy
        and a plane defined by ax + by + cz + d = 0
        """
        x0 = self.interp_lines[0]((indx, indy))
        y0 = self.interp_lines[1]((indx, indy))
        z0 = self.interp_lines[2]((indx, indy))
        dx = self.interp_lines[3]((indx, indy))
        dy = self.interp_lines[4]((indx, indy))
        dz = self.interp_lines[5]((indx, indy))
        # fmt: off
        t = -(a*x0 + b*y0 + c*z0 + d) / (a*dx + b*dy + c*dz)
        # fmt: on
        physical_coords = np.array([x0 + t * dx, y0 + t * dy, z0 + t * dz])
        return physical_coords.transpose()

    def apply_calib(self, indx, indy, dx, dy, a, b, c, d):
        """Gives the projection of the real displacement projected on each
        camera plane and then projected on the laser plane
        """
        displacements = self.intersect_with_plane(
            indx + dx / 2.0, indy + dy / 2.0, a, b, c, d
        ) - self.intersect_with_plane(
            indx - dx / 2.0, indy - dy / 2.0, a, b, c, d
        )
        return displacements

    def get_base_camera_plane(self, indx=None, indy=None):
        """Matrix of base change from camera plane to fixed plane"""
        if indx is None:
            indx = range(self.nb_pixels_x // 2 - 20, self.nb_pixels_x // 2 + 20)
            indy = range(self.nb_pixels_y // 2 - 20, self.nb_pixels_y // 2 + 20)
            indx, indy = np.meshgrid(indx, indy)
        dx = np.nanmean(self.interp_lines[3]((indx, indy)))
        dy = np.nanmean(self.interp_lines[4]((indx, indy)))
        dz = np.nanmean(self.interp_lines[5]((indx, indy)))
        A, B = get_base_from_normal_vector(dx, dy, dz)
        return A, B

    def check_interp_levels(self, number=100):
        """Plot to check interp_levels"""
        interp = self.interp_levels
        indx = range(0, self.nb_pixels_x, self.nb_pixels_x // number)
        indy = range(0, self.nb_pixels_y, self.nb_pixels_y // number)
        indx, indy = np.meshgrid(indx, indy)
        Z = interp.Z
        for i in range(len(Z)):
            X = interp.indices_pixel2xphys[i]((indx, indy))
            Y = interp.indices_pixel2yphys[i]((indx, indy))
            pylab.figure()
            pylab.pcolor(indx, indy, X, shading="nearest")
            pylab.title(f"Level {i}, X")
            pylab.xlabel("x (pix)")
            pylab.ylabel("y (pix)")
            pylab.colorbar()

            pylab.figure()
            pylab.pcolor(indx, indy, Y, shading="nearest")
            pylab.title(f"Level {i}, Y")
            pylab.xlabel("x (pix)")
            pylab.ylabel("y (pix)")
            pylab.colorbar()

        plt.show()

    def check_interp_lines(self, number=10):
        """Plot to check interp_lines"""
        fig = pylab.figure()
        ax = Axes3D(fig, auto_add_to_figure=False)
        fig.add_axes(ax)

        for i, path in enumerate(self.paths_xml):
            X, Y, Z, x, y = get_points(path)
            ax.scatter(X, Y, Z, marker=".", color="blue")

        x = range(0, self.nb_pixels_x, self.nb_pixels_x // number)
        y = range(0, self.nb_pixels_y, self.nb_pixels_y // number)
        x, y = np.meshgrid(x, y)
        x = x.flatten()
        y = y.flatten()
        for i in range(len(x)):
            X0 = self.interp_lines[0]((x[i], y[i]))
            Y0 = self.interp_lines[1]((x[i], y[i]))
            Z0 = self.interp_lines[2]((x[i], y[i]))
            dx = self.interp_lines[3]((x[i], y[i]))
            dy = self.interp_lines[4]((x[i], y[i]))
            dz = self.interp_lines[5]((x[i], y[i]))
            X = (np.arange(10) - 5) / 20.0 * dx + X0
            Y = (np.arange(10) - 5) / 20.0 * dy + Y0
            Z = (np.arange(10) - 5) / 20.0 * dz + Z0
            ax.plot(X, Y, Z, "r")
        ax.set_xlabel("x (m)")
        ax.set_ylabel("y (m)")
        ax.set_zlabel("z (m)")
        plt.show()

    def check_interp_lines_coeffs(self, number=100):
        """Plot to check interp_lines coefficients"""
        x = range(0, self.nb_pixels_x, self.nb_pixels_x // number)
        y = range(0, self.nb_pixels_y, self.nb_pixels_y // number)
        x, y = np.meshgrid(x, y)

        X0 = np.zeros(x.shape)
        Y0 = np.zeros(x.shape)
        Z0 = np.zeros(x.shape)
        dx = np.zeros(x.shape)
        dy = np.zeros(x.shape)
        dz = np.zeros(x.shape)

        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                X0[i, j] = self.interp_lines[0]((x[i, j], y[i, j]))
                Y0[i, j] = self.interp_lines[1]((x[i, j], y[i, j]))
                Z0[i, j] = self.interp_lines[2]((x[i, j], y[i, j]))
                dx[i, j] = self.interp_lines[3]((x[i, j], y[i, j]))
                dy[i, j] = self.interp_lines[4]((x[i, j], y[i, j]))
                dz[i, j] = self.interp_lines[5]((x[i, j], y[i, j]))

        pylab.figure()
        pylab.pcolor(x, y, X0, shading="nearest")
        pylab.title("X0")
        pylab.xlabel("x (pix)")
        pylab.ylabel("y (pix)")
        pylab.colorbar()
        pylab.figure()
        pylab.pcolor(x, y, Y0, shading="nearest")
        pylab.title("Y0")
        pylab.xlabel("x (pix)")
        pylab.ylabel("y (pix)")
        pylab.colorbar()
        pylab.figure()
        pylab.pcolor(x, y, Z0, shading="nearest")
        pylab.title("Z0")
        pylab.xlabel("x (pix)")
        pylab.ylabel("y (pix)")
        pylab.colorbar()
        pylab.figure()
        pylab.pcolor(x, y, dx, shading="nearest")
        pylab.title("dx")
        pylab.xlabel("x (pix)")
        pylab.ylabel("y (pix)")
        pylab.colorbar()
        pylab.figure()
        pylab.pcolor(x, y, dy, shading="nearest")
        pylab.title("dy")
        pylab.xlabel("x (pix)")
        pylab.ylabel("y (pix)")
        pylab.colorbar()
        pylab.figure()
        pylab.pcolor(x, y, dz, shading="nearest")
        pylab.title("dz")
        pylab.xlabel("x (pix)")
        pylab.ylabel("y (pix)")
        pylab.colorbar()
        pylab.figure()
        pylab.pcolor(
            x, y, np.sqrt(dx**2 + dy**2 + dz**2), shading="nearest"
        )
        pylab.title("norm(d)")
        pylab.xlabel("x (pix)")
        pylab.ylabel("y (pix)")
        pylab.colorbar()

        plt.show()


class DirectStereoReconstruction:
    """Class to get stereo reconstruction with direct Calibration
    This calibration determine the equations of optical paths for "each" pixels

    Parameters
    ----------

    path_file0:
       Path of the file of the first camera

    path_file1:
       Path of the file of the second camera
    """

    def __init__(self, path_file0, path_file1):
        self.calib0 = CalibDirect(path_file=path_file0)
        self.calib1 = CalibDirect(path_file=path_file1)
        # matrices from camera planes to fixed plane and inverse
        self.A0, self.B0 = self.calib0.get_base_camera_plane()
        self.A1, self.B1 = self.calib1.get_base_camera_plane()

        if np.allclose(self.A0, self.A1):
            raise ValueError("The two calibrations have to be different.")

        # M1, M2: see reconstruction function
        self.invM0 = np.linalg.inv(np.vstack([self.B0[0:2, :], self.B1[0:1, :]]))
        self.invM1 = np.linalg.inv(np.vstack([self.B0[0:1, :], self.B1[0:2, :]]))
        self.invM2 = np.linalg.inv(np.vstack([self.B0[0:2, :], self.B1[1:2, :]]))
        self.invM3 = np.linalg.inv(np.vstack([self.B0[1:2, :], self.B1[0:2, :]]))

    def project2cam(
        self,
        indx0,
        indy0,
        dx0,
        dy0,
        indx1,
        indy1,
        dx1,
        dy1,
        a,
        b,
        c,
        d,
        check=False,
    ):
        """Project displacements of each cameras dx0, dy0, dx1 and dy1
        in their respective planes.
        """

        X0 = self.calib0.intersect_with_plane(indx0, indy0, a, b, c, d)
        dX0 = self.calib0.apply_calib(indx0, indy0, dx0, dy0, a, b, c, d)

        X1 = self.calib1.intersect_with_plane(indx1, indy1, a, b, c, d)
        dX1 = self.calib1.apply_calib(indx1, indy1, dx1, dy1, a, b, c, d)

        d0cam = np.tensordot(self.B0, dX0.swapaxes(0, 1), axes=1)[
            :2, :
        ].transpose()
        d1cam = np.tensordot(self.B1, dX1.swapaxes(0, 1), axes=1)[
            :2, :
        ].transpose()

        if check:
            plt.figure()
            plt.quiver(indx0, indy0, dx0, dy0)
            plt.axis("equal")
            pylab.xlabel("x (pix)")
            pylab.ylabel("y (pix)")
            plt.title("Cam 0 in pixel")
            plt.figure()
            plt.quiver(X0[:, 0], X0[:, 1], dX0[:, 0], dX0[:, 1])
            pylab.xlabel("x (pix)")
            pylab.ylabel("y (pix)")
            plt.axis("equal")
            plt.title("Cam 0 in m")
            plt.figure()
            plt.quiver(indx1, indy1, dx1, dy1)
            pylab.xlabel("x (pix)")
            pylab.ylabel("y (pix)")
            plt.axis("equal")
            plt.title("Cam 1 in pixel")
            plt.figure()
            plt.quiver(X1[:, 0], X1[:, 1], dX1[:, 0], dX1[:, 1])
            pylab.xlabel("x (pix)")
            pylab.ylabel("y (pix)")
            plt.axis("equal")
            plt.title("Cam 1 in m")
            plt.show()

        return X0, X1, d0cam, d1cam

    def find_common_grid(self, X0, X1, a, b, c, d):
        """Find a common grid for the 2 cameras"""
        xmin0 = np.nanmin(X0[:, 0])
        xmax0 = np.nanmax(X0[:, 0])
        ymin0 = np.nanmin(X0[:, 1])
        ymax0 = np.nanmax(X0[:, 1])
        xmin1 = np.nanmin(X1[:, 0])
        xmax1 = np.nanmax(X1[:, 0])
        ymin1 = np.nanmin(X1[:, 1])
        ymax1 = np.nanmax(X1[:, 1])

        xmin = max([xmin0, xmin1])
        xmax = min([xmax0, xmax1])
        ymin = max([ymin0, ymin1])
        ymax = min([ymax1, ymax1])

        Lx0 = xmax0 - xmin0
        Ly0 = ymax0 - ymin0
        Lx1 = xmax1 - xmin1
        Ly1 = ymax1 - ymin1

        Nx0 = sqrt(X0[~np.isnan(X0[:, 0]), 0].size) * sqrt(Lx0 / Ly0)
        Ny0 = X0[~np.isnan(X0[:, 1]), 1].size / Nx0
        Nx1 = sqrt(X1[~np.isnan(X1[:, 0]), 0].size) * sqrt(Lx1 / Ly1)
        Ny1 = X1[~np.isnan(X1[:, 1]), 1].size / Nx1

        dx0 = Lx0 / Nx0
        dy0 = Ly0 / Ny0
        dx1 = Lx1 / Nx1
        dy1 = Ly1 / Ny1

        dx = max([dx0, dx1])
        dy = max([dy0, dy1])

        x = np.linspace(xmin, xmax, int((xmax - xmin) / dx))
        y = np.linspace(ymin, ymax, int((ymax - ymin) / dy))
        x, y = np.meshgrid(x, y)
        self.grid_x = x.transpose()
        self.grid_y = y.transpose()
        self.grid_z = -(a * self.grid_x + b * self.grid_y + d) / c
        return self.grid_x, self.grid_y, self.grid_z

    def interp_on_common_grid(self, X0, X1, d0cam, d1cam, grid_x, grid_y):
        """Interpolate displacements of the 2 cameras d0cam, d1cam on the
        common grid grid_x, grid_y
        """
        # if not hasattr(self, 'grid_x'):
        #     self.find_common_grid(X0, X1, a, b, c, d)
        ind0 = (
            (~np.isnan(X0[:, 0]))
            * (~np.isnan(X0[:, 1]))
            * (~np.isnan(d0cam[:, 0]))
            * (~np.isnan(d0cam[:, 1]))
        )
        ind1 = (
            (~np.isnan(X1[:, 0]))
            * (~np.isnan(X1[:, 1]))
            * (~np.isnan(d1cam[:, 0]))
            * (~np.isnan(d1cam[:, 1]))
        )

        d0xcam = griddata(
            (X0[ind0, 0], X0[ind0, 1]), d0cam[ind0, 0], (grid_x, grid_y)
        )
        d0ycam = griddata(
            (X0[ind0, 0], X0[ind0, 1]), d0cam[ind0, 1], (grid_x, grid_y)
        )
        d1xcam = griddata(
            (X1[ind1, 0], X1[ind1, 1]), d1cam[ind1, 0], (grid_x, grid_y)
        )
        d1ycam = griddata(
            (X1[ind1, 0], X1[ind1, 1]), d1cam[ind1, 1], (grid_x, grid_y)
        )
        return d0xcam, d0ycam, d1xcam, d1ycam

    def reconstruction(
        self, X0, X1, d0cam, d1cam, a, b, c, d, grid_x, grid_y, check=False
    ):
        """Reconstruction of the 3 components of the velocity in the plane
        defined by a, b, c, d on the grid defined by grid_x, grid_y
        from the displacements of the 2 cameras d0cam, d1cam in their respective
        planes d0cam, d1cam
        """
        # reconstruction => resolution of the equation
        # MX = dcam

        d0xcam, d0ycam, d1xcam, d1ycam = self.interp_on_common_grid(
            X0, X1, d0cam, d1cam, grid_x, grid_y
        )

        # in the case where 2 vectors from different cameras are approximately
        # the same, they don't have to be used both in the computation
        n1 = np.abs(np.inner(self.A0[0], self.A1[0]))
        n2 = np.abs(np.inner(self.A0[1], self.A1[1]))
        n3 = np.abs(np.inner(self.A0[1], self.A1[0]))
        n4 = np.abs(np.inner(self.A0[0], self.A1[1]))
        # I suppose that 5deg between vectors is sufficient
        threshold = np.cos(5 * 2 * np.pi / 180.0)

        tmp = []
        dcam = np.zeros((3, d0xcam.shape[0], d0xcam.shape[1]))
        if n1 < threshold and n3 < threshold:
            dcam[0, :, :] = d0xcam
            dcam[1, :, :] = d0ycam
            dcam[2, :, :] = d1xcam
            tmp.append(np.tensordot(self.invM0, dcam, axes=([1], [0])))
        if n1 < threshold and n4 < threshold:
            dcam[0, :, :] = d0xcam
            dcam[1, :, :] = d1xcam
            dcam[2, :, :] = d1ycam
            tmp.append(np.tensordot(self.invM1, dcam, axes=([1], [0])))
        if n2 < threshold and n4 < threshold:
            dcam[0, :, :] = d0xcam
            dcam[1, :, :] = d0ycam
            dcam[2, :, :] = d1ycam
            tmp.append(np.tensordot(self.invM2, dcam, axes=([1], [0])))
        if n2 < threshold and n3 < threshold:
            dcam[0, :, :] = d0ycam
            dcam[1, :, :] = d1xcam
            dcam[2, :, :] = d1ycam
            tmp.append(np.tensordot(self.invM3, dcam, axes=([1], [0])))
        tmp = np.array(tmp)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            tmp_mean = np.nanmean(tmp, 0)
        tmp_std = np.nanstd(tmp, 0)
        dX, dY, dZ = tmp_mean
        Errorx, Errory, Errorz = tmp_std

        # dXx = np.zeros(d0xcam.shape)
        # dXy = np.zeros(d0xcam.shape)
        # dXz = np.zeros(d0xcam.shape)
        # Errorx = np.zeros(d0xcam.shape)
        # Errory = np.zeros(d0xcam.shape)
        # Errorz = np.zeros(d0xcam.shape)
        # for i in range(d0xcam.shape[0]):
        #     for j in range(d0xcam.shape[1]):
        #         tmp = []
        #         if n1 < threshold and n3 < threshold:
        #             dcam = np.hstack(
        #                 [d0xcam[i, j], d0ycam[i, j], d1xcam[i, j]])
        #             tmp.append(np.dot(self.invM0, dcam))
        #         if n1 < threshold and n4 < threshold:
        #             dcam = np.hstack(
        #                 [d0xcam[i, j], d1xcam[i, j], d1ycam[i, j]])
        #             tmp.append(np.dot(self.invM1, dcam))
        #         if n2 < threshold and n4 < threshold:
        #             dcam = np.hstack(
        #                 [d0xcam[i, j], d0ycam[i, j], d1ycam[i, j]])
        #             tmp.append(np.dot(self.invM2, dcam))
        #         if n2 < threshold and n3 < threshold:
        #             dcam = np.hstack(
        #                 [d0ycam[i, j], d1xcam[i, j], d1ycam[i, j]])
        #             tmp.append(np.dot(self.invM3, dcam))
        #         tmp = np.array(tmp)

        # dXx[i, j] = np.nanmean(tmp, 0)[0]
        # Errorx[i, j] = np.nanstd(tmp, 0)[0]
        # dXy[i, j] = np.nanmean(tmp, 0)[1]
        # Errory[i, j] = np.nanstd(tmp, 0)[1]
        # dXz[i, j] = np.nanmean(tmp, 0)[2]
        # Errorz[i, j] = np.nanstd(tmp, 0)[2]

        if check:
            plt.figure()
            plt.quiver(self.grid_x, self.grid_y, d0xcam, d0ycam)
            pylab.xlabel("X")
            pylab.ylabel("Y")
            plt.axis("equal")
            plt.title("Cam 0 plane projection")
            plt.figure()
            plt.quiver(self.grid_x, self.grid_y, d1xcam, d1ycam)
            pylab.xlabel("X")
            pylab.ylabel("Y")
            plt.axis("equal")
            plt.title("Cam 1 plane projection")
            plt.figure()
            plt.quiver(self.grid_x, self.grid_y, dX, dY)
            pylab.xlabel("X")
            pylab.ylabel("Y")
            plt.axis("equal")
            plt.title("Reconstruction on laser plane")
            plt.show()

        return dX, dY, dZ, Errorx, Errory, Errorz



import unittest
from shutil import rmtree

import numpy as np
import h5py

from fluiddyn.util.paramcontainer import ParamContainer
from fluidimage import get_path_image_samples
from fluidimage.postproc.piv import ArrayPIV, PIV2d, get_grid_pixel_from_piv_file


class TestPIV(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path_piv = get_path_image_samples() / "Milestone/piv_0000a-b.h5"
        cls.path_tests = cls.path_piv.parent / "tmp_tests"
        cls.path_tests.mkdir(exist_ok=True)

    @classmethod
    def tearDownClass(cls):
        if cls.path_tests.exists():
            rmtree(cls.path_tests)

    def test_get_grid(self):
        xs1d, ys1d = get_grid_pixel_from_piv_file(self.path_piv)

    def test_from_MultipassPIVResults(self):
        PIV2d.from_file(self.path_piv)

    def test_piv_objects(self):

        ny, nx = 8, 12

        xs1d = np.linspace(-1, 1, nx)
        ys1d = np.linspace(-1, 1, ny)
        xs2d, ys2d = np.meshgrid(xs1d, ys1d)
        piv0 = PIV2d(xs1d, ys1d, 0, xs2d, ys2d)

        piv0.display()
        piv0.extract(0, 0.5, 0, 0.8, phys=True)

        piv1 = (2 * piv0 + 1) / 3 + piv0

        piv1.compute_spatial_fft()
        piv1.compute_divh()
        piv1.compute_rotz()
        piv1.compute_norm()

        arr = ArrayPIV((piv0,))
        arr.append(piv1)
        arr.extend(
            (piv0 - 1, 1 + piv1 * piv0 - piv0 / piv0 - 1, 0 + piv0 * 3 + 1)
        )

        arr.set_timestep(0.1)
        assert len(arr.times) == 5
        assert arr.times.max() == 0.4

        arr.compute_temporal_fft()

        arr1 = 0 + 2 * arr * 3 + 1 - 2
        arr1 = (1 + arr1 / 2).median_filter(3).gaussian_filter(0.5)

        piv2 = arr1[0]
        arr1[1] = piv2
        repr(arr1)
        len(arr1)
        del arr1[-1]

        arr1 = arr1.truncate().extract(0, 4, 2, 7).extract_square()

        with h5py.File(self.path_piv, "r") as file:
            params = ParamContainer(hdf5_object=file["params"])

        path = self.path_tests / "piv2d.h5"
        piv2.save(path, params=params)
        piv_loaded = PIV2d.from_file(path, load_params=True)

        assert np.allclose(piv2.x, piv_loaded.x)
        assert np.allclose(piv2.vx, piv_loaded.vx)
        assert piv2.namevx == piv_loaded.namevx
        assert piv2.unitvx == piv_loaded.unitvx
        assert params == piv_loaded.params

    def test_compute_time_average(self):

        ny, nx = 8, 12

        xs1d = np.linspace(-1, 1, nx)
        ys1d = np.linspace(-1, 1, ny)
        xs2d, ys2d = np.meshgrid(xs1d, ys1d)
        piv = PIV2d(xs1d, ys1d, 0, xs2d, ys2d)
        nt = 5

        arr = ArrayPIV([piv] * nt)

        piv_mean = arr.compute_time_average()

        assert np.allclose(piv.vx, piv_mean.vx)
        assert np.allclose(piv.vy, piv_mean.vy)


if __name__ == "__main__":
    unittest.main()



from time import clock

import matplotlib.pyplot as plt
import numpy as np
import scipy as sc
from matplotlib import mlab
from scipy import interpolate

from fluidimage.calcul.interpolate import thin_plate_spline_subdom
from fluidimage.calcul.interpolate.griddata import griddata
from fluidimage.calcul.interpolate.thin_plate_spline import (
    ThinPlateSpline,
    compute_tps_coeff,
)

plt.ion()


n0 = 40
n1 = 100

method = "griddata_scipy"
method = "griddata_mpl"
method = "griddata_fluidimage"
# method = 'tps'
# method = 'tps_subdom'

with_gradient = False


def myplot(i, x, y, U, title=None):
    plt.figure(i)
    ax = plt.gca()

    ax.scatter(x, y, c=U, vmin=-1, vmax=1)

    if title is not None:
        ax.set_title(title)


L = 2 * np.pi

x = y = np.linspace(0, L, n0)
x, y = np.meshgrid(x, y)
x = x.ravel()
y = y.ravel()

U = np.exp(-((x - np.pi) ** 2 + (y - np.pi) ** 2))  # gaussian
# U = x  # linear

myplot(0, x, y, U, "input data")

centers = np.vstack([x, y])
# interpolation grid
xI = yI = np.linspace(0, L, n1)
XI, YI = np.meshgrid(xI, yI)
XI = XI.ravel()
YI = YI.ravel()

dx = L / n1

new_positions = np.vstack([XI, YI])

t0 = clock()

if method == "tps":

    # calculate tps coeff
    U_smooth, U_tps = compute_tps_coeff(centers, U, 0)
    # evaluate interpolation on the new grid
    tps = ThinPlateSpline(new_positions, centers)

    U_eval = tps.compute_field(U_tps)

    if with_gradient:
        DUX_eval, DUY_eval = tps.compute_gradient(U_tps)

elif method == "tps_subdom":

    smoothing_coef = 0.5
    subdom_size = 50

    tps = thin_plate_spline_subdom.ThinPlateSplineSubdom(
        centers, subdom_size, smoothing_coef, threshold=1, pourc_buffer_area=0.5
    )

    U_smooth, U_tps = tps.compute_tps_coeff_subdom(U)

    tps.init_with_new_positions(new_positions)
    U_eval = tps.compute_eval(U_tps)

elif method == "griddata_mpl":

    U_eval = mlab.griddata(x, y, U, xI, yI, "linear")

elif method == "griddata_scipy":

    grid_x, grid_y = np.meshgrid(xI, yI)
    # nearest, linear, cubic
    U_eval = interpolate.griddata(centers.T, U, (grid_x, grid_y), "cubic", 0)

elif method == "griddata_fluidimage":
    # grid_x, grid_y = np.meshgrid(xI, yI)
    U_eval = griddata(centers, U, (xI, yI))
else:
    raise ValueError


if method.startswith("griddata") and with_gradient:
    DUY_eval, DUX_eval = sc.gradient(U_eval, dx, dx)

print("done in {} s ({})".format(clock() - t0, method))

myplot(1, XI, YI, U_eval, "U_eval")

if with_gradient:
    myplot(2, XI, YI, DUX_eval, "DUX_eval")
    myplot(3, XI, YI, DUY_eval, "DUY_eval")

plt.show()



import logging
import unittest

import numpy as np

from fluidimage.calcul.correl import (
    CorrelFFTBase,
    CorrelPyCuda,
    CorrelPythran,
    CorrelScipySignal,
    CorrelTheano,
    correlation_classes,
)
from fluidimage.synthetic import make_synthetic_images

# config_logging('debug')
logger = logging.getLogger("fluidimage")

classes = {k.replace(".", "_"): v for k, v in correlation_classes.items()}
classes2 = {
    "sig": CorrelScipySignal,
    "theano": CorrelTheano,
    "pycuda": CorrelPyCuda,
    "pythran": CorrelPythran,
}

try:
    from reikna.cluda import any_api

    api = any_api()
except Exception:
    classes.pop("cufft")

try:
    import pycuda
except ImportError:
    classes.pop("pycuda")
    classes2.pop("pycuda")

try:
    import skcuda
except ImportError:
    classes.pop("skcufft")

try:
    import theano
except ImportError:
    classes.pop("theano")
    classes2.pop("theano")


class TestCorrel(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        nx = 32
        ny = 32

        # Sometimes errors with nx = ny = 16
        # nx = 16
        # ny = 16

        displacement_x = 3.3
        displacement_y = 5.8

        cls.displacements = np.array([displacement_x, displacement_y])

        nb_particles = (nx // 4) ** 2

        cls.im0, cls.im1 = make_synthetic_images(
            cls.displacements, nb_particles, shape_im0=(ny, nx), epsilon=0.0
        )


for k, cls in classes.items():

    def _test(self, cls=cls, k=k):

        if issubclass(cls, CorrelFFTBase):
            displacement_max = "50%"
        else:
            displacement_max = None

        correl = cls(
            self.im0.shape, self.im1.shape, displacement_max=displacement_max
        )

        # first, no displacement
        c, norm = correl(self.im0, self.im0)
        (
            dx,
            dy,
            correl_max,
            other_peaks,
        ) = correl.compute_displacements_from_correl(c, norm=norm)
        displacement_computed = np.array([dx, dy])

        logger.debug(
            k
            + ", displacement = [0, 0]\t error= {}\n".format(
                abs(displacement_computed)
            )
        )

        self.assertTrue(np.allclose([0, 0], displacement_computed, atol=1e-03))

        # then, with the 2 figures with displacements
        c, norm = correl(self.im0, self.im1)
        dx, dy, correl_max, _ = correl.compute_displacements_from_correl(
            c, norm=norm
        )

        displacement_computed = np.array([dx, dy])

        logger.debug(
            k
            + ", displacement = {}\t error= {}\n".format(
                self.displacements,
                abs(displacement_computed - self.displacements),
            )
        )

        self.assertTrue(
            np.allclose(self.displacements, displacement_computed, atol=0.8)
        )

    exec("TestCorrel.test_correl_square_image_" + k + " = _test")


class TestCorrel1(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        nx = 32
        ny = 64

        # Sometimes errors with nx = ny = 16
        # nx = 16
        # ny = 16

        displacement_x = 3.3
        displacement_y = 5.8

        cls.displacements = np.array([displacement_x, displacement_y])

        nb_particles = (max(nx, ny) // 4) ** 2

        cls.im0, cls.im1 = make_synthetic_images(
            cls.displacements, nb_particles, shape_im0=(ny, nx), epsilon=0.0
        )


for k, cls in classes.items():

    def _test1(self, cls=cls, k=k):
        correl = cls(self.im0.shape, self.im1.shape)

        # first, no displacement
        c, norm = correl(self.im0, self.im0)
        dx, dy, correl_max, _ = correl.compute_displacements_from_correl(
            c, norm=norm
        )
        displacement_computed = np.array([dx, dy])

        logger.debug(
            k
            + ", displacement = [0, 0]\t error= {}\n".format(
                abs(displacement_computed)
            )
        )

        self.assertTrue(np.allclose([0, 0], displacement_computed, atol=1e-03))

        # then, with the 2 figures with displacements
        c, norm = correl(self.im0, self.im1)
        dx, dy, correl_max, _ = correl.compute_displacements_from_correl(
            c, norm=norm
        )

        displacement_computed = np.array([dx, dy])

        logger.debug(
            k
            + ", displacement = {}\t error= {}\n".format(
                self.displacements,
                abs(displacement_computed - self.displacements),
            )
        )

        self.assertTrue(
            np.allclose(self.displacements, displacement_computed, atol=0.8)
        )

    exec("TestCorrel1.test_correl_rectangular_image_" + k + " = _test1")


class TestCorrel2(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        nx0 = 48
        ny0 = 96
        nx1 = 32
        ny1 = 64

        # Sometimes errors with nx = ny = 16
        # nx = 16
        # ny = 16

        displacement_x = 3.3
        displacement_y = 5.8

        cls.displacements = np.array([displacement_x, displacement_y])

        nb_particles = (max(nx0, ny0) // 4) ** 2

        cls.im0, cls.im1 = make_synthetic_images(
            cls.displacements,
            nb_particles,
            shape_im0=(ny0, nx0),
            shape_im1=(ny1, nx1),
            epsilon=0.0,
        )

        cls.im1 = cls.im1.astype("float32")


for k, cls in classes2.items():

    def _test2(self, cls=cls, k=k):
        correl = cls(self.im0.shape, self.im1.shape, mode="valid")

        # with the 2 figures with displacements
        c, norm = correl(self.im0, self.im1)
        dx, dy, correl_max, _ = correl.compute_displacements_from_correl(
            c, norm=norm
        )

        displacement_computed = np.array([dx, dy])

        logger.debug(
            k
            + ", displacement = {}\t error= {}\n".format(
                self.displacements,
                abs(displacement_computed - self.displacements),
            )
        )

        self.assertTrue(
            np.allclose(self.displacements, displacement_computed, atol=0.8)
        )

    exec("TestCorrel2.test_correl_images_diff_sizes" + k + " = _test2")

if __name__ == "__main__":
    unittest.main()



"""Preprocessing data objects (:mod:`fluidimage.data_objects.preproc`)
======================================================================

.. autoclass:: ArraySerie
   :members:
   :private-members:

.. autoclass:: PreprocResults
   :members:
   :private-members:

"""

import math
import os

import h5py

from fluiddyn.util.serieofarrays import SerieOfArraysFromFiles
from fluidimage.util import imsave, imsave_h5

from .piv import ArrayCouple, LightPIVResults


def get_Ni_Nj(serie):
    """Returns number of images in the first and second indices
    of the series.

    """
    if not isinstance(serie, SerieOfArraysFromFiles):
        raise ValueError(
            "serie must be an instance of class SerieOfArraysFromFiles"
        )

    nb_indices = serie.nb_indices
    slices = serie.get_index_slices()
    Ni = slices[0][1] - slices[0][0]
    if nb_indices == 1:
        Nj = 1
    elif nb_indices == 2:
        if len(slices[1]) == 1:
            Nj = 1
        else:
            Nj = slices[1][1] - slices[1][0]
    else:
        raise NotImplementedError(
            "Cannot evaluate series with more than 2 indices"
        )

    return Ni, Nj


def get_ind_middle(serie):
    Ni, Nj = get_Ni_Nj(serie)
    ind_middle_start = int(math.floor(Ni / 2.0)) * Nj
    ind_middle_end = int(math.ceil(Ni / 2.0)) * Nj
    if Ni % 2 == 0:
        ind_middle_start -= 1

    return ind_middle_start, ind_middle_end


def get_name_preproc(serie, name_files, ind_serie, nb_series, out_format):
    ind_middle_start, ind_middle_end = get_ind_middle(serie)

    if ind_serie == 0 and nb_series == 1:
        s = -1
    elif ind_serie == 0:
        s = ind_middle_start
    elif ind_serie == nb_series - 1:
        s = -1
    else:
        s = ind_middle_start

    if out_format == "img":
        return name_files[s]

    else:
        fname, ext = os.path.splitext(name_files[s])
        fname += "." + out_format
        return fname


class ArraySerie(ArrayCouple):
    def __init__(
        self,
        names=None,
        arrays=None,
        serie=None,
        ind_serie=0,
        nb_series=1,
        str_path=None,
        hdf5_parent=None,
    ):

        if str_path is not None:
            self._load(path=str_path)
            return

        if hdf5_parent is not None:
            self._load(hdf5_object=hdf5_parent["serie"])
            return

        if serie is not None:
            names = serie.get_name_arrays()
            paths = serie.get_path_arrays()
            self.paths = tuple(os.path.abspath(p) for p in paths)

            if arrays is None:
                arrays = serie.get_arrays()

        self.ind_serie = ind_serie
        self.nb_series = nb_series
        self.names = tuple(names)
        self.arrays = tuple(arrays)
        self.serie = serie

    def _clear_data(self):
        self.arrays = tuple()

    def save(self, path=None, hdf5_parent=None):
        if path is not None:
            raise NotImplementedError

        if not isinstance(hdf5_parent, (h5py.File,)):
            raise NotImplementedError

        hdf5_parent.create_group("serie")
        group = hdf5_parent["serie"]
        group.attrs["names"] = self.names
        group.attrs["paths"] = self.paths


class PreprocResults(LightPIVResults):
    def __init__(self, params=None, str_path=None, hdf5_object=None):

        self._keys_to_be_saved = ["data"]
        if hdf5_object is not None:
            if params is not None:
                self.params = params

            self._load_from_hdf5_object(hdf5_object)
            return

        if str_path is not None:
            self._load(str_path)
            return

        self.params = params
        self.data = {}

    def _clear_data(self):
        self.data = {}

    def save(self, path=None):
        out_format = self.params.saving.format
        for k, v in self.data.items():
            path_file = os.path.join(path, k)
            if out_format == "img":
                imsave(path_file, v, as_int=True)
            elif out_format == "h5":
                attrs = {
                    "class_name": "PreprocResults",
                    "module_name": self.__module__,
                }
                imsave_h5(path_file, v, self.params, attrs, as_int=True)
            else:
                # Try to save in formats supported by PIL.Image
                imsave(path_file, v, format=out_format, as_int=True)

        self._clear_data()
        return self



import numpy as np

from .thin_plate_spline import (
    ThinPlateSpline,
    ThinPlateSplineNumpy,
    compute_tps_coeff,
)

pi = np.pi


def test_tps():
    n0 = 10
    n1 = 10

    # set of random x coordinates from 0 to 2pi
    x = 2 * pi * np.random.rand(n0)
    y = 2 * pi * np.random.rand(n0)

    U = np.exp(-((x - pi) ** 2 + (y - pi) ** 2))  # gaussian

    centers = np.vstack([x, y])
    # interpolation grid
    xI = yI = np.linspace(0, 2 * pi, n1)
    XI, YI = np.meshgrid(xI, yI)
    XI = XI.ravel()
    YI = YI.ravel()

    new_positions = np.vstack([XI, YI])

    # calculate tps coeff
    U_smooth, U_tps = compute_tps_coeff(centers, U, 0)
    # evaluate interpolation on the new grid
    tps = ThinPlateSpline(new_positions, centers)
    tps.compute_field(U_tps)
    tps.compute_gradient(U_tps)

    ThinPlateSplineNumpy(new_positions, centers)



"""Thin plate spline with subdomains
====================================

Translated and adapted from UVmat code (Joel Sommeria, LEGI, CNRS).

This interpolation/smoothing (Duchon, 1976; NguyenDuc and Sommeria,
1988) minimises a linear combination of the squared curvature and
squared difference from the initial data.

We first need to compute tps coefficients ``U_tps`` (function
``compute_tps_coeff``). Interpolated data can then be obtained as the
matrix product ``dot(U_tps, EM)`` where the matrix ``EM`` is obtained
by the function ``compute_tps_matrix``.  The spatial derivatives are
obtained as ``dot(U_tps, EMDX)`` and ``dot(U_tps, EMDY)``, where
``EMDX`` and ``EMDY`` are obtained from the function
``compute_tps_matrix_dxy``. A helper class is also provided.

.. autoclass:: ThinPlateSplineSubdom
   :members:

"""

from logging import debug

import numpy as np

from .thin_plate_spline import compute_tps_coeff, compute_tps_matrix


class ThinPlateSplineSubdom:
    """Helper class for thin plate interpolation."""

    def __init__(
        self,
        centers,
        subdom_size,
        smoothing_coef,
        threshold=None,
        percent_buffer_area=0.2,
    ):

        self.centers = centers
        self.subdom_size = subdom_size
        self.smoothing_coef = smoothing_coef
        self.threshold = threshold
        self.compute_indices(percent_buffer_area)

    def compute_indices(self, percent_buffer_area=0.25):
        xs = self.centers[1]
        ys = self.centers[0]
        max_coord = np.max(self.centers, 1)
        min_coord = np.min(self.centers, 1)
        range_coord = max_coord - min_coord
        aspect_ratio = range_coord[0] / range_coord[1]

        nb_subdom = xs.size / self.subdom_size
        nb_subdomx = int(np.floor(np.sqrt(nb_subdom / aspect_ratio)))
        nb_subdomx = nb_subdomx or 1
        nb_subdomy = int(np.ceil(np.sqrt(nb_subdom * aspect_ratio)))
        nb_subdomy = nb_subdomy or 1

        debug(f"nb_subdomx: {nb_subdomx} ; nb_subdomy: {nb_subdomy}")

        nb_subdom = nb_subdomx * nb_subdomy

        self.nb_subdomx = nb_subdomx
        self.nb_subdomy = nb_subdomy

        x_dom = np.linspace(min_coord[1], max_coord[1], nb_subdomx + 1)
        y_dom = np.linspace(min_coord[0], max_coord[0], nb_subdomy + 1)

        buffer_area_x = (
            range_coord[1]
            / (nb_subdomx)
            * percent_buffer_area
            * np.ones_like(x_dom)
        )
        buffer_area_y = (
            range_coord[0]
            / (nb_subdomy)
            * percent_buffer_area
            * np.ones_like(y_dom)
        )

        self.x_dom = x_dom
        self.y_dom = y_dom
        self.buffer_area_x = buffer_area_x
        self.buffer_area_y = buffer_area_y

        ind_subdom = np.zeros([nb_subdom, 2])
        ind_v_subdom = []

        i_subdom = 0
        for i in range(nb_subdomx):
            for j in range(nb_subdomy):
                ind_subdom[i_subdom, :] = [i, j]

                ind_v_subdom.append(
                    np.where(
                        (xs >= x_dom[i] - buffer_area_x[i])
                        & (xs < x_dom[i + 1] + buffer_area_x[i + 1])
                        & (ys >= y_dom[j] - buffer_area_y[j])
                        & (ys < y_dom[j + 1] + buffer_area_y[j + 1])
                    )[0]
                )

                i_subdom += 1
        self.ind_v_subdom = ind_v_subdom
        self.nb_subdom = nb_subdom

    def compute_tps_coeff_subdom(self, U):

        U_smooth = [None] * self.nb_subdom
        U_tps = [None] * self.nb_subdom

        for i in range(self.nb_subdom):

            centers_tmp = self.centers[:, self.ind_v_subdom[i]]

            U_tmp = U[self.ind_v_subdom[i]]
            U_smooth[i], U_tps[i] = self.compute_tps_coeff_iter(
                centers_tmp, U_tmp
            )

        U_smooth_tmp = np.zeros(self.centers[1].shape)
        nb_tps = np.zeros(self.centers[1].shape, dtype=int)

        for i in range(self.nb_subdom):
            U_smooth_tmp[self.ind_v_subdom[i]] += U_smooth[i]
            nb_tps[self.ind_v_subdom[i]] += 1

        U_smooth_tmp /= nb_tps

        return U_smooth_tmp, U_tps

    def init_with_new_positions(self, new_positions):
        npos = self.new_positions = new_positions

        ind_new_positions_subdom = []

        x_dom = self.x_dom
        y_dom = self.y_dom
        buffer_area_x = self.buffer_area_x
        buffer_area_y = self.buffer_area_y

        i_subdom = 0
        for i in range(self.nb_subdomx):
            for j in range(self.nb_subdomy):
                ind_new_positions_subdom.append(
                    np.where(
                        (npos[1] >= x_dom[i] - buffer_area_x[i])
                        & (npos[1] < x_dom[i + 1] + buffer_area_x[i + 1])
                        & (npos[0] >= y_dom[j] - buffer_area_y[j])
                        & (npos[0] < y_dom[j + 1] + buffer_area_y[j + 1])
                    )[0]
                )

                i_subdom += 1

        self.ind_new_positions_subdom = ind_new_positions_subdom
        self._init_EM_subdom()

    def _init_EM_subdom(self):

        EM = [None] * self.nb_subdom

        for i in range(self.nb_subdom):
            centers_tmp = self.centers[:, self.ind_v_subdom[i]]
            new_positions_tmp = self.new_positions[
                :, self.ind_new_positions_subdom[i]
            ]
            EM[i] = compute_tps_matrix(new_positions_tmp, centers_tmp)

        self.EM = EM

    def compute_eval(self, U_tps):

        U_eval = np.zeros(self.new_positions[1].shape)
        nb_tps = np.zeros(self.new_positions[1].shape, dtype=int)

        for i in range(self.nb_subdom):
            U_eval[self.ind_new_positions_subdom[i]] += np.dot(
                U_tps[i], self.EM[i]
            )
            nb_tps[self.ind_new_positions_subdom[i]] += 1

        U_eval /= nb_tps

        return U_eval

    def compute_tps_coeff_iter(self, centers, U):
        """Compute the thin plate spline (tps) coefficients removing erratic
        vectors

        It computes iteratively "compute_tps_coeff", compares the tps
        result to the initial data and remove it if difference is
        larger than the given threshold

        """
        U_smooth, U_tps = compute_tps_coeff(centers, U, self.smoothing_coef)
        if self.threshold is not None:
            Udiff = np.sqrt((U_smooth - U) ** 2)
            ind_erratic_vector = np.argwhere(Udiff > self.threshold)

            count = 1
            while ind_erratic_vector.size != 0:
                U[ind_erratic_vector] = U_smooth[ind_erratic_vector]
                U_smooth, U_tps = compute_tps_coeff(
                    centers, U, self.smoothing_coef
                )

                Udiff = np.sqrt((U_smooth - U) ** 2)
                ind_erratic_vector = np.argwhere(Udiff > self.threshold)
                count += 1

                if count > 10:
                    print("tps stopped after 10 iterations")
                    break

        if count > 1:
            print("tps done after ", count, " attempt(s)")
        return U_smooth, U_tps
